from __future__ import annotations

from os import error
from typing import Any

from simudator.core.signal import Signal


class Module:
    """
    This class specifies the basic functionality of all module types.
    The modules are processor components in the processor,
    such as ALU, busses, registers and more.

    Parameters
    ----------
    signals : dict[str, Signal]
        Dictionary containing all signals connected to the module. Input
        signals should have keys beginning with 'in\_' and output signals keys
        beginning with 'out\_'.
    name : str
        Name of the module.

    Attributes
    ----------
    name : str
        Name of the module.
    signals : dict[str, Signal]
        The signals connected to this module. The key for a signal can be
        regarded as the name of a signal 'port' of the module. The keys should
        be prefixed with 'in\_' for input signals and 'out\_' for output signals.
    """

    __slots__ = ("name", "signals")

    def __init__(self, signals: dict[str, Signal], name: str = "") -> None:
        self.name = name
        self.signals = signals

        for key, signal in self.signals.items():
            if key[0:2] == "in":
                signal.add_destination(self)

    def update_register(self) -> None:
        """Simulate module behaviour for saving input into the internal state
        during a clock tick.

        This method should be used for any module that has any 'internal'
        registers. Exact behaviour is specified in each module subclass.
        """
        raise NotImplementedError

    def output_register(self) -> None:
        """Simulate module behaviour for outputting data from internal
        registers, if any, to output signals during a clock tick.

        Exact behaviour is specified in each module subclass.
        """
        raise NotImplementedError

    def update_logic(self) -> None:
        """
        Simulate behaviour for sending output using the internal state
        during a clock tick. Exact behaviour is specified in each module type.
        """
        raise NotImplementedError

    def get_state(self) -> dict[str, Any]:
        """
        Return the state of the module.

        Returns
        -------
        dict[str, Any]
            State of the module represented as a dictionary with one key for
            each state variable.
        """
        state_dict = dict()
        state_dict["name"] = self.name
        return state_dict

    def get_parameter(self) -> dict[str, Any]:
        """
        Return the parameters of the module.

        Parameters are immutable attributes of modules that affect how they operate,
        such as the bit length of internal registers. They are set once during the initialization
        of the modules and assumed to be constant throughout the execution of the program.

        Returns
        -------
        dict[str, Any]
            Parameters of the module represented as a dictionary with one key for
            each parameter variable.
        """
        param_dict = dict()
        return param_dict

    def get_gui_state(self) -> dict:
        """
        Return the state of the module as should be displayed in a GUI.

        See Also
        --------
        get_state :
            Similar method that includes state variables not related to the GUI.
        """
        state_dict = dict()
        state_dict["name"] = self.name
        return state_dict

    def set_state(self, state: dict) -> None:
        """
        Set the state of the module.

        Parameters
        ----------
        state : dict[str, Any]
            Module state represented as a dictionary.
        """
        self.name = state["name"]

    def get_output_signals(self) -> list[Signal]:
        """Return the output signals of the module.

        Assumes all output signals are stored with a key beginning in 'out\_' in
        the signals dictionary.

        Returns
        -------
        list[Signal]
            List of all output signals of this module.
        """
        return [signal for key, signal in self.signals.items() if key[0:3] == "out"]

    def get_input_signals(self) -> list[Signal]:
        """Return the input signals of the module.

        Assumes all input signals are stored with a key beginning in 'in\_' in
        the signals dictionary.

        Returns
        -------
        list[Signal]
            List of all input signals of this module.
        """
        return [signal for key, signal in self.signals.items() if key[0:2] == "in"]

    def reset(self) -> None:
        """Reset the module to its default state."""
        raise NotImplementedError

    def load_from_str(self, state_string) -> None:
        """Set the module state according to a string.

        Each module subclass decides the format of the string.
        """
        raise NotImplementedError

    def get_longest_line_len(self, ignore_keys=None) -> int:
        """Return the length of the longest line when printing the module.

        Helper function for pretty_print.
        """

        if ignore_keys == None:
            ignore_keys = []

        longest_line_len = 0
        state = self.get_state()

        for key in state:
            if key not in ignore_keys:
                line = key + ": " + str(state[key])
                if len(line) > longest_line_len:
                    longest_line_len = len(line)

        return longest_line_len

    def print_module(self) -> None:
        """Print the module."""
        print(self.name)

    def _helper_save_state_to_file(self, file_path: str, content: str) -> bool:
        """Save the given content to a given file.

        Parameters
        ----------
        file_path : str
            Path to file to save to.
        content : str
            Content to save to file.

        Returns
        -------
        ``True`` on success, ``False`` otherwise.
        """
        try:
            with open(file_path, "a") as file:
                file.write(content)
        except OSError:
            return False
        return True

    def save_state_to_file(self, file_path: str) -> bool:
        """Save the modules state to file.

        Parameters
        ----------
        file_path : str
            Path to the file to save to.

        Returns
        -------
        bool
            True if saving succeeded, False otherwise.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Module({self.name!r})"
