from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Memory(Module):
    """
    A general size memory module.

    Parameters
    ----------
    input : Signal
        Signal of which the value may be written to the current address.
    output : Signal
        Signal to which the content of the current address is written.
    control : Signal
        Signal used to enable or disable write to the current address. Assumed
        to output ``bool``.
    address : Signal
        Signal used to address the memory. Assumed to output ``int``.
    size : int
        Size of the address space of the memory.
    name : str
        Name of the module.
    """

    __slots__ = (
        "_memory",
        "_current_address",
        "_is_write",
    )
    """
    Constants used to textually pad the values of the memory when saving to
    file.
    """
    MEMORY_VALUE_PADDING = 2
    MEMORY_ADDRESS_PADDING = 2

    def __init__(
        self,
        input: Signal,
        output: Signal,
        control: Signal,
        address: Signal,
        size: int,
        name="Memory",
    ) -> None:

        # signals
        signals = {
            "in_input": input,
            "in_control": control,
            "in_address": address,
            "out_content": output,
        }

        # Init super class
        super().__init__(signals, name)

        # Internal state
        self._memory = [0 for _ in range(size)]
        self._current_address = 0
        self._is_write = False

    def update_register(self):
        if self._is_write:
            value = self.signals["in_input"].get_value()
            self._memory[self._current_address] = value

    def output_register(self) -> None:
        pass
        # Might need to output the current value here
        # in case the memory was written to in update_register()

    def update_logic(self) -> None:
        adr_sig = self.signals["in_address"]
        ctrl_sig = self.signals["in_control"]
        out_sig = self.signals["out_content"]
        if adr_sig.get_value() is not None and ctrl_sig.get_value() is not None:
            self._is_write = ctrl_sig.get_value()
            self._current_address = adr_sig.get_value()
            out_sig.update_value(self._memory[self._current_address])

    def print_module(self) -> None:
        print("", self.name, "\n -----")
        for address, value in enumerate(self._memory):
            if value:
                print("", str(hex(address)), "", str(hex(value)), "\n")

    def get_state(self) -> dict[str, Any]:
        state = super().get_state()
        state["is_write"] = self._is_write
        state["current_address"] = self._current_address
        state["memory"] = self._memory[:]
        return state

    def get_gui_state(self) -> dict[str, Any]:
        state = super().get_gui_state()
        state["current_address"] = self._current_address
        state["memory"] = self._memory[:]
        return state

    def set_state(self, state: dict[str, Any]) -> None:
        super().set_state(state)
        if "is_write" in state:
            self._is_write = state["is_write"]
        if "current_address" in state:
            self._current_address = state["current_address"]
        if "memory" in state:
            self._memory = state["memory"]

    def reset(self) -> None:
        """
        Reset the memory to 0 for each address.
        """
        for i in range(len(self._memory)):
            self._memory[i] = 0

    def get_longest_line_len(self, ignore_keys=None) -> int:
        """
        Helper function for pretty_print that returns the length of
        the longest value in the memory to print for a module.
        """

        if ignore_keys is None:
            ignore_keys = []

        longest_memory_line = 0

        for value in self._memory:
            value_len = len(str(value))
            if value_len > longest_memory_line:
                longest_memory_line = value_len

        return longest_memory_line

    def get_largest_mem_adr(self) -> int:
        """
        Helper function for pretty_print that returns the length of
        the largest address in the memory to print for a module.
        """
        return len(str(len(self._memory)))

    def save_state_to_file(self, file_path: str) -> bool:
        content = ""
        content = self.name + ":\n"
        for index, value in enumerate(self._memory):
            content += (
                str(hex(index)[2:].rjust(self.MEMORY_ADDRESS_PADDING, "0")) + ": "
            )
            content += str(hex(value)[2:].rjust(self.MEMORY_VALUE_PADDING, "0"))
            content += "\n"
        content += "\n"
        return super()._helper_save_state_to_file(file_path, content)
