from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Register(Module):
    """
    A simple module that can store and output a value.

    Parameters
    ----------
    input : Signal
        Signal from which the value is stored in the register.
    output : Signal
        Signal onto which the value of the register is outputted.
    value : Any
        Initial value of the register.
    name : str
        Name of the register.
    """

    __slots__ = "_value"

    def __init__(
        self,
        input: Signal,
        output: Signal,
        value: Any = 0,
        name: str = "Register",
    ) -> None:

        # signals
        signals = {
            "in_content": input,
            "out_content": output,
        }

        # init the instance
        super().__init__(signals, name)
        self._value = value

    def update_register(self) -> None:
        """Update the internal value of the register with the input signal.

        Any extra bits that do not fit the bit length of the register are
        thrown away.
        """
        input_value = self.signals["in_content"].get_value()
        if input_value is None:
            return
        self._value = input_value

    def output_register(self) -> None:
        """Output the value of the register onto the output signal."""
        self.signals["out_content"].update_value(self._value)

    def update_logic(self):
        """
        Do nothing.

        The register has no logic.
        """
        pass

    def get_state(self) -> dict[str, Any]:
        state = super().get_state()
        state["value"] = self._value
        return state

    def get_gui_state(self) -> dict[str, Any]:
        state = self.get_state()
        return state

    def set_state(self, state: dict[str, Any]) -> None:
        """
        Set the state of the register.

        Parameters
        ----------
        state : dict[str, Any]
            The state of the register to load. Should contain the keys "name"
            and "value" with values of type ``str`` and ``int`` respectively.
        """
        self.name = state["name"]
        self._value = state["value"]

    def reset(self) -> None:
        """
        Reset the register to 0.
        """
        self._value = 0

    def save_state_to_file(self, file_path: str) -> bool:
        content = self.name + ":\n" + "value: " + str(self._value) + "\n\n"
        return super()._helper_save_state_to_file(file_path, content)

    def print_module(self) -> None:
        print(
            "",
            self.name,
            "\n -----",
            "\n value: ",
            self._value,
        )


class IntegerRegister(Register):
    """
    A register intended to store integers only.

    Parameters
    ----------
    input : Signal
        Signal from which the value is stored in the register.
    output : Signal
        Signal onto which the value of the register is outputted.
    bit_length : int
        Maximum number of bits of the input to store in the register. All extra
        bits of the input value are discarded.
    value : Any
        Initial value of the register.
    name : str
        Name of the register.
    """

    __slots__ = "_bit_length"

    def __init__(
        self,
        input: Signal,
        output: Signal,
        bit_length: int,
        value: int = 0,
        name: str | None = None,
    ) -> None:

        # set the register name
        if name is None:
            name = f"{bit_length}-bit register"

        super().__init__(input, output, value=value, name=name)

        # set the bit length of the register
        self._bit_length = bit_length

    def update_register(self) -> None:
        super().update_register()
        mask = 2**self._bit_length - 1
        self._value = self._value & mask

    def get_state(self) -> dict[str, Any]:
        state = super().get_state()
        return state

    def get_parameter(self) -> dict[str, Any]:
        """
        Return a dict of the register parameter, i.e. its bit length.
        """
        parameter = super().get_parameter()
        parameter["bit_length"] = self._bit_length
        return parameter

    def set_state(self, state: dict[str, Any]) -> None:
        super().set_state(state)

    def save_state_to_file(self, file_path: str) -> None:
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + hex(self._value)[2:] + "\n\n")
        file.close()

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        self._value = int(string_pair[1], 16)


class Flag(IntegerRegister):
    """
    A simple module that can store a value. This module is also
    updated during the 'update logic' phase.
    """

    def __init__(
        self,
        input: Signal,
        output: Signal,
        bit_length=1,
        value=0,
        name="Flag",
    ) -> None:

        # set the flags name
        super().__init__(
            input=input,
            output=output,
            bit_length=bit_length,
            value=value,
            name=name,
        )

    def update_logic(self):
        """
        The flag has no logic but it still need to be updated during
        the update logic phase. This is due the fact that the flag
        should update the same tick as the change that causes the flag
        to activate (ex: loop counter reaches zero -> set L flag to 1)
        """
        input_value = self.signals["in_content"].get_value()
        mask = 2**self._bit_length - 1
        self._value = input_value & mask

    def get_gui_state(self) -> dict:
        state = super().get_gui_state()
        state["value"] = self._value
        return state

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("value: " + str(self._value) + "\n\n")
        file.close()

    def print_module(self) -> None:
        print(
            "",
            self.name,
            "\n -----",
            "\n value: ",
            self._value,
            "\n bit length: ",
            self._bit_length,
        )
