import os
from collections.abc import Callable
from typing import Any

from simudator.core.breakpoint import Breakpoint
from simudator.core.breakpoint_lambda import LambdaBreakpoint
from simudator.core.breakpoint_memory import MemoryBreakpoint
from simudator.core.breakpoint_state import StateBreakpoint
from simudator.core.module import Module

from .signal import Signal


class Processor:
    """
    Main class for controlling the processor.

    Uses modules and signals to simulate processor components behaviour.

    Attributes
    ----------
    is_stopped : bool
        Used to indicate that the processor has stopped, e.g. from a halt
        instruction.
    breakpoint_reached : bool
        Used to indicate that the processor has reached a breakpoint at
        the current clock cycle.
    last_breakpoint : Breakpoint | None
        Stores the last reached breakpoint if any.

    """

    __slots__ = (
        "_modules",
        "_signals",
        "_clock",
        "_update_queue",
        "_module_history",
        "_signal_history",
        "_breakpoint_id_counter",
        "_breakpoints",
        "breakpoint_reached",
        "last_breakpoint",
        "_removed_cycles",
        "is_stopped",
        "_new_instruction",
        "_current_instructions",
        "_ignore_keys",
        "_lambdas",
        "_assembly_cycles",
    )

    LINE_SEPARATOR = "-"
    MAX_LINE_LEN = 170
    CYCLES_TO_SAVE = 10000

    def __init__(self) -> None:
        # For simulation
        self._modules: dict[str, Module] = dict()
        self._signals: list[Signal] = []
        self._clock: int = 0
        self._update_queue: list[Module] = []
        self.is_stopped: bool = False

        # Breakpoint handling
        self._breakpoint_id_counter: int = 1
        self._breakpoints: dict[int, Breakpoint] = {}
        self.breakpoint_reached: bool = False
        self.last_breakpoint: Breakpoint | None = None
        self._lambdas: dict[str, Callable[..., bool]] = {}

        # Saving processor state for saving/loading to file and undoing ticks
        self._removed_cycles: int = 0
        self._assembly_cycles: list[int] = [0]  # Map asm instruction to clock cycle
        self._module_history: list[dict[str, dict[str, Any]]] = []
        self._signal_history: list[list] = []  # TODO: Is this needed?

        # For showing which instructions are being done and signalling
        # the start of a new one
        self._new_instruction: bool = False

        # It is the responsibility of the CPU to tell the gui:s pipeline
        # diagram what information to be displayed and how to display it.
        # Thus each CPU keeps track of its current instructions together with
        # each instructions position in the pipeline diagram.
        self._current_instructions: list[tuple[str, int, int]] = []

        # TODO: keeping track of what pieces of info not to show
        # show not be done at the processor level.
        # Maybe implement a 'get_pretty_print_state' at module
        # level?
        self._ignore_keys: list[str] = [
            "bit_length",
            "mask",
        ]

    def do_tick(self) -> None:
        """Simulate one clock cycle of the processor.

        Run each module's tick function, then handle the following updates.
        Also check for breakpoints that are reached in this cycle, and save
        clock cycles when new assembly instructions are started.
        """
        if len(self._module_history) > self._clock - self._removed_cycles:
            # If a previous stored cycle has been loaded, discard
            # all stored cycles from that cycle and onward in
            # the history of saved cycles
            self._module_history = self._module_history[0 : self._clock]
            self._signal_history = self._signal_history[0 : self._clock]

        self.unstop()
        self.save_cycle()
        self._clock += 1

        for module in self._modules.values():
            module.update_register()

        for module in self._modules.values():
            module.output_register()

        while self._update_queue:
            module = self._update_queue.pop(0)
            module.update_logic()

        self.stop_at_breakpoints()
        if self.should_halt():
            self.stop()

        # Set new_instruction and save clock cycle for each
        # new assembly instruction
        if self.is_new_instruction():
            self._new_instruction = True
            self._current_instructions = self.get_current_instructions()
            self._assembly_cycles.append(self.get_clock())
        else:
            self._new_instruction = False

    def get_current_instructions(self) -> list[tuple[str, int, int]]:
        """Return a list of the current instructions with their positions.

        Each entry in the list is a tuple of a instruction with its column and
        row position for the pipeline diagram.

        Returns
        -------
        list[tuple[str, int, int]]
            A list of instructions represented as tuples. Each tuple contains
            the instruction and its position in the pipeline.
        """
        raise NotImplementedError

    def get_pipeline_dimensions(self) -> tuple[int, int]:
        """Return the dimensions of the pipeline of the processor.

        Returns
        -------
        width : int
            The width of the pipeline (number of cycles to display at a time).
        depth : int
           Depth of the pipeline.
        """
        raise NotImplementedError

    def get_asm_instruction(self) -> str:
        """Return a string containing the 'code' for the current assembly
        instruction.

        What 'code' refers to is up to each processor to define. It could be a
        memory address, an instruction label or something else.

        Returns
        -------
        str
            The current assembly instruction.
        """
        raise NotImplementedError

    def run_asm_instruction(self, num_instructions=1) -> None:
        """Run a number of assembly instructions.

        Defaults to running 1 instruction unless some other number is provided.
        Assumes ``is_new_instruction()`` is implemented.

        Parameters
        ----------
        num_instructions : int
            Number of assembly instructions to perform.

        """
        while num_instructions > 0:

            self.do_tick()

            if self.is_new_instruction():
                num_instructions -= 1

    def undo_asm_instruction(self, num_instructions=1) -> None:
        """
        Undo a number of assembly instructions on the CPU.

        Default argument is one, but it is possible to specify any number of
        instructions.

        Parameters
        ----------
        num_instructions : int
            Number of assembly instructions to undo.
        """

        current_clock_cycle = self._clock
        index = len(self._assembly_cycles)
        saved_cycle = self._assembly_cycles[index - 1]

        # Make sure we are undoing the instruction(s) we are currently on
        while saved_cycle >= current_clock_cycle:
            index -= 1
            saved_cycle = self._assembly_cycles[index - 1]

        index -= num_instructions

        # We cant undo more instructions than we have performed.
        if index < 0:
            raise IndexError

        clockcycle = self._assembly_cycles[index]

        self.load_cycle(clockcycle)

        # Need +1 here since we save the start state to enable to
        # load the start state. This is done since we only append clock
        # cycles to the list self.assembly_cycles when we reach a new state
        # that has uPC set to 0, which wont happen when we load a new file.
        self._assembly_cycles = self._assembly_cycles[: index + 1]

    def run_continuously(self) -> None:
        """
        Run the processor until it halts, is stopped or reaches a breakpoint.
        """
        self.unstop()
        while not self.is_stopped:
            self.do_tick()

    def should_halt(self) -> bool:
        """
        Return whether the processor should halt this clock cycle or not.

        Returns
        -------
        bool
            ``True`` if the processor should halt, ``False`` otherwise.

        Notes
        ----
        Each processor should implement the condition for True in its subclass.
        If we instead raise NotImplementedError as we do in all other cases
        when the functionality should be implemented in the subclass we lose
        the ability to test modules on the base processor class.
        """
        return False

    def stop(self) -> None:
        """Signal to stop the execution of the processor until the processor is
        instructed to run continuously or do some ticks again.
        """
        self.is_stopped = True

    def unstop(self) -> None:
        """Reset the stop execution signal."""
        self.is_stopped = False

    def is_new_instruction(self) -> bool:
        """
        Return true on the same clock cycle the CPU starts a new instruction.

        Should be implemented per CPU, return field new_instruction. It is up
        to each processor to set this field to True/False correctly.
        """
        return self._new_instruction

    def stop_at_breakpoints(self) -> None:
        """Stop the execution if any breakpoint has been reached during this
        clock cycle.

        Also record the breakpoint that was reached.
        """
        self.breakpoint_reached = False
        for _, bp in self._breakpoints.items():
            # TODO: Can make this more efficient by only checking enabled
            # breakpoints
            if bp.is_break() and bp.is_enabled:
                self.breakpoint_reached = True
                self.last_breakpoint = bp

        if self.breakpoint_reached:
            self.stop()

    def reset(self):
        """Reset the processor to its start state.

        This resets all modules and removes any saved states for undoing clock
        cycles. A round of value propagation is done to reset signals too.
        """
        self._clock = 0
        self._module_history.clear()
        self._signal_history.clear()
        self._removed_cycles = 0
        self._assembly_cycles = [0]

        for module in self._modules.values():
            module.reset()
            module.output_register()

        while self._update_queue:
            module = self._update_queue.pop(0)
            module.update_logic()

    def add_modules_to_update(self, module: Module) -> None:
        """Queue a module to be updated at the end of clock cycle.

        Parameters
        ----------
        module : Module
            Module to queue for update.

        See Also
        --------
        Module.update_logic :
            Method for updating a module.
        """
        if module not in self._update_queue:
            self._update_queue.append(module)

    def add_module(self, module: Module) -> None:
        """Add module to be simulated by the processor.

        Parameters
        ----------
        module : Module
            Module to add.
        """
        self._modules[module.name] = module

    def get_module(self, name: str) -> Module:
        """Get module with specific name.

        Parameters
        ----------
        name : str
            Name of the module to get.

        Returns
        -------
        Module
            The module with the specified name.
        """
        return self._modules[name]

    def get_modules(self) -> list[Module]:
        """Get list of all modules.

        Returns
        -------
        list[Module]
            List of all modules in the processor.
        """
        return list(self._modules.values())

    def add_signals(self, signals: list[Signal]) -> None:
        """Add signals to the processor.

        Parameters
        ----------
        list[Signal]
            List of signals to add for simulation.
        """
        self._signals += signals

    def get_clock(self) -> int:
        """Get the current clockcycle number.

        Returns
        -------
        int
            Current clock cycle number of the processor.
        """
        return self._clock

    def set_clock(self, value: int) -> None:
        """Set current clockcycle number.

        Parameters
        ----------
        value : int
            Cycle number to set the clock to.
        """
        self._clock = value

    def save_cycle(self) -> None:
        """
        Save the current state for all the modules and signals in the
        processor to the processor's internal history of all clock cycles.
        """
        module_states = {}

        # Only save a specified number of cycles,
        # saving every cycle can easily eat all ram
        if len(self._module_history) > Processor.CYCLES_TO_SAVE:
            self._module_history.pop(0)
            self._removed_cycles += 1

        for module in self._modules.values():
            module_states[module.name] = module.get_state()
        self._module_history.append(module_states)

    def load_cycle(self, cycle: int) -> None:
        """Load the state of all modules as they were at the specified clock
        cycle.

        This does not erase the history of states for the later cycles.

        Parameters
        ----------
        cycle : int
            Number of the cycle to load.
        """

        cycle_index = cycle - self._removed_cycles

        if cycle_index < 0:
            raise ValueError("The cycle to be loaded is not saved")

        try:
            module_states = self._module_history[cycle_index]
        except IndexError:
            raise IndexError

        for module_name, module_state in module_states.items():
            self._modules[module_name].set_state(module_state)

        self._clock = cycle

        for module in self._modules.values():
            module.output_register()

        while self._update_queue:
            module = self._update_queue.pop(0)
            module.update_logic()

    def load_state_from_file(self, file_path: str) -> None:
        """Load states for modules from a file.

        Parameters
        ----------
        file_path : str
            Path to the file containing the processor state to load.
        """
        self.reset()

        file = open(file_path)

        # Ensure we are at the start of the file
        file.seek(0, os.SEEK_END)  # Go to end of file
        if file.tell():  # If current pos != 0
            file.seek(0)  # Return to the start of the file
        else:
            raise ValueError  # Else raise error

        # Appends the lines from the loaded file to a long string. This string
        # is given to each module in the function 'load_from_str'. Each module
        # will thus load itself.
        module_name = None
        module_str = ""

        # Go through each line in the file
        lines = file.readlines()
        for line in lines:

            # If no name currently saved then get name from current line
            if module_name is None:
                module_name = line.partition(":")[0]

            # if already has name get line content
            else:
                # if not empty line, then save content
                if line.strip():
                    module_str += line

                # if line is empty then give saved content to name file
                else:
                    module = self.get_module(module_name)
                    module.load_from_str(module_str)
                    module_str = ""
                    module_name = None

        # The last module will be an edge case, since we append the information
        # after we read an empty line.
        if module_name and module_str:
            module = self.get_module(module_name)
            module.load_from_str(module_str)

    def save_state_to_file(self, file_path: str) -> bool:
        """Save the current state of the cpu to file.

        Will overwrite the given file.

        Parameters
        ----------
        file_path : str
            Path to the file to write to.
        """
        file = open(file_path, "w")
        file.write("")
        file.close()

        for module in self._modules.values():
            res = module.save_state_to_file(file_path)

            if not res:
                return False

        return True

    def get_breakpoints(self) -> dict[int, Breakpoint]:
        """
        Get the breakpoints in the processor as a map from breakpoint id to
        breakpoint.

        Returns
        -------
        dict[int, Breakpoint]
            Map from breakpoint id to breakpoint, containing all breakpoints
            in the processor.
        """
        return self._breakpoints

    def add_state_breakpoint(
        self, module_name: str, state_name: str, value: Any
    ) -> None:
        """Add a module state breakpoint to the processor.

        Parameters
        ----------
            module_name : str
                Name of the module for which a breakpoint will be added.
            state_name : str
                Name of the state variable of the module to put a breakpoint
                on.
            value : Any
                Value of the state variable that triggers the breakpoint.

        Raises
        ------
        ValueError
            If the module does not exist or if the state variable does not
            exist in the module.
        """

        if module_name not in self._modules:
            raise ValueError(f"No module named {module_name}")

        module = self._modules[module_name]

        if state_name not in module.get_state():
            raise ValueError(f"No state named {state_name} in " f"module {module_name}")

        bp = StateBreakpoint(module, state_name, value)
        self._breakpoints[self._breakpoint_id_counter] = bp
        self._breakpoint_id_counter += 1

    def remove_breakpoint(self, id: int) -> bool:
        """Remove a specific breakpoint.

        Parameters
        ----------
        id : int
            ID of the breakpoint to be removed.
        """
        if id in self._breakpoints:
            del self._breakpoints[id]
            return True
        return False

    def print_breakpoints(self) -> None:
        """Print all breakpoints of the processor."""
        if not self._breakpoints:
            print("There are no breakpoints.")
        else:
            for bp_id, bp in self._breakpoints.items():
                print(f"BP {bp_id}: {bp}")

    def add_memory_breakpoint(self, module_name: str, address: int, value: Any) -> None:
        """Add a breakpoint for a memory module.

        Parameters
        ----------
        module_name : str
            Name of the memory module for which a breakpoint will be added.
        address : int
            Address of the memory module to break at.
        value : Any
            Value of the address that triggers the breakpoint.

        Raises
        ------
        ValueError
            If the module does not exist or is not a memory.
        """

        if module_name not in self._modules:
            raise ValueError(f"No module named {module_name}")

        module = self._modules[module_name]
        if "memory" not in module.get_state():
            raise ValueError(f"Module {module_name} is not a memory.")

        bp = MemoryBreakpoint(module, address, value)
        self._breakpoints[self._breakpoint_id_counter] = bp
        self._breakpoint_id_counter += 1

    def get_breakpoint_lambdas(self) -> list[str]:
        """Get all functions available for lambda breakpoints.

        Returns
        -------
        list[str]
            List of names of all functions of the processor that are available
            for adding lambda breakpoints.
        """

        return list(self._lambdas.keys())

    def add_lambda_breakpoint(self, lambda_name: str, **kwargs) -> None:
        """Add a lambda breakpoint to the processor.

        Module names used in the keyword arguments are substituted by the
        actual modules before being passed on to the lambda breakpoint
        function.

        Parameters
        ----------
        lambda_name : str
            Name of the lambda breakpoint function to be used by the
            breakpoint.
        **kwargs
            Arguments to pass to the breakpoint function.

        Raises
        ------
        ValueError
            If the  lambda breakpoint function does not exist.
        """
        # Substitute any module names for the module instances
        # This is to be able to externally create general lambda breakpoints
        # that can take any values as arguments while also having a way to take
        # modules as arguments without having direct access to them
        for key in kwargs.keys():
            value = kwargs[key]
            if value in self._modules:
                kwargs[key] = self._modules[value]

        if lambda_name not in self._lambdas:
            raise ValueError(f"No lambda named {lambda_name}.")

        lambda_func = self._lambdas[lambda_name]
        bp = LambdaBreakpoint(lambda_func, **kwargs)
        self._breakpoints[self._breakpoint_id_counter] = bp
        self._breakpoint_id_counter += 1

    def set_enabled_breakpoint(self, bp_id: int, is_enabled: bool) -> None:
        """Toggle a breakpoint to enabled or disabled.

        Parameters
        ----------
        bp_id : int
            ID of the breakpoint to toggle.
        is_enabled : bool
            ``True`` to enable, ``False`` to disable the breakpoint.
        """
        self._breakpoints[bp_id].set_enabled(is_enabled)
