import json
from json.decoder import JSONDecodeError

from qtpy.QtCore import QPointF
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import (
    QErrorMessage,
    QFileDialog,
    QGraphicsItem,
    QGraphicsScene,
    QMessageBox,
)

from simudator.core.processor import Processor
from simudator.gui.color_scheme import ColorScheme
from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem
from simudator.gui.signal_graphics_item import SignalGraphicsItem


class CpuGraphicsScene(QGraphicsScene):
    """
    This class creates the graphics scene for visualising the processor.
    It takes each module representation as a ModuleGraphicsItem
    and handles mouse inputs for interacting with these graphicsitems.
    Can create a default layout on creation and can save/load new layouts.
    """

    MODULE_SPACEING = 100

    def __init__(self):
        super().__init__()
        self._module_graphics_items = dict()
        self._signal_graphics_items = []
        self._error_msg_box = QErrorMessage()
        self.setBackgroundBrush(ColorScheme.Window)

    def reset_signals(self) -> None:
        """
        Reset all graphical signals to their default visual representation
        when initialised.
        """
        for graphics_signal in self._signal_graphics_items:
            graphics_signal.reset()

    def add_module(self, item: ModuleGraphicsItem) -> None:
        """
        Add a graphical module item to the processor scene at a position
        according to the default layout.

        Parameters
        ----------
        item : ModuleGraphicsItem
            Graphical module item to add to the scene.
        """
        self._module_graphics_items[item.name] = item
        self._place_module_default(item)

    def _place_module_default(self, item: ModuleGraphicsItem) -> None:
        """
        Place a graphical module item at a position according to the default
        layout, i.e. placing all module items along a diagonal.

        Parameters
        ----------
        item : ModuleGraphicsItem
        """
        placement = len(self._module_graphics_items) * self.MODULE_SPACEING
        item.setPos(placement, placement)
        self.addItem(item)

    def move_module_to(self, item: ModuleGraphicsItem, pos_x: int, pos_y: int) -> None:
        """
        Place an existing graphical module item at some position.

        Parameters
        ----------
        item : ModuleGraphicsItem
           Graphical module item to place at a position.
        pos_x : int
            Position on x-axis to place the module item at.
        pos_y : int
            Position on y-axis to place the module item at.
        """
        item.setPos(pos_x * self.MODULE_SPACEING, pos_y * self.MODULE_SPACEING)

    def add_all_signals(self) -> None:
        """
        Instantiate graphical signals between all matching ports of all
        graphical modules.
        """
        # Map which ports that are connected using their signals' names
        # (Every port keeps a reference to a simulation signal which has
        # a name)
        signal_to_ports = {}
        for module_w in self._module_graphics_items.values():
            ports = module_w.getPorts()
            for port in ports:
                s_name = port.getSignalName()
                if s_name not in signal_to_ports:
                    signal_to_ports[s_name] = [port]
                else:
                    signal_to_ports[s_name].append(port)

        # Create a signal graphics_item for each bundle of connected ports
        for module_graphics_items in signal_to_ports.values():
            # TODO: Implement signals that go to multiple destinations
            if len(module_graphics_items) == 2:
                port_1 = module_graphics_items[0]
                port_2 = module_graphics_items[1]
                signal_w = SignalGraphicsItem(port_1, port_2)
                self._signal_graphics_items.append(signal_w)
                self.addItem(signal_w)
                port_1.moved.connect(signal_w.move)
                port_2.moved.connect(signal_w.move)
                port_1.toggled.connect(signal_w.toggleVisibility)
                port_2.toggled.connect(signal_w.toggleVisibility)

    def get_modules(self) -> list[ModuleGraphicsItem]:
        """
        Get a list of all graphical module items in the scene.

        Returns
        -------
        list[ModuleGraphicsItem]
            List of graphical module items in the scene.
        """
        return list(self._module_graphics_items.values())

    def get_signals(self) -> list[SignalGraphicsItem]:
        """
        Get a list of all graphical signal items in the scene.

        Returns
        -------
        list[SignalGraphicsItem]
            List of graphical signal items in the scene.
        """
        return self._signal_graphics_items

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

    @Slot()
    def load_layout(self) -> None:
        """
        Prompt the user for a layout file and load the layout from the file.
        """

        # Prompt the user for a file
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        dialog.setDirectory("~/simudator")  # TODO: does this work when exported?
        path = dialog.getOpenFileName()[0]

        # If no file was selected, do nothing
        if path == '':
            return

        try:
            self.load_layout_from_file(path)

        # Anything goes wrong with loading the selected one,
        # we dont care about what went wrong
        except (OSError, JSONDecodeError):
            self.load_default_layout()
            self._error_msg_box.showMessage(
                "Unable to load given file.", "load_layout_err"
            )
        else:
            QMessageBox.information(
                self.parent(), "SimuDator", "File loaded succesfully."
            )

    def _save_layout_to_file(self, file_path: str) -> None:
        """
        Save the layout of the scene to file.

        Parameters
        ----------
        file_path : str
            Path to the file to save to layout to.
        """
        layout_str = ""
        for graphics_item in self._module_graphics_items.values():
            layout_str += graphics_item.save_state_as_str() + "\n"

        file = open(file_path, "w")
        file.write(layout_str)
        file.close()

    @Slot(bool)
    def show_all_signals(self, value: bool) -> None:
        """
        Set the visibility of all graphical signals in the scene.

        Parameters
        ----------
        value : bool
            ``True`` to show all graphical signals, ``False`` to hide them.
        """
        for item in self._signal_graphics_items:
            item.setVisible(value)

    @Slot(bool)
    def show_port_names(self, value: bool) -> None:
        """
        Set the visibility of the names of all ports of graphical modules
        in the scene.

        Parameters
        ----------
        value : bool
            ``True`` to show the port names, ``False`` to hide them.
        """
        for item in self._module_graphics_items.values():
            for port in item.ports:
                port.setNameVisibility(value)

    def toggle_layout_lock(self, value: bool) -> None:
        """
        Toggle the layout between locked and unlocked. Locked means that
        nothing can be moved around in the layout.

        Parameters
        ----------
        value : bool
            ``True`` to lock the layout, ``False`` to unlock it.
        """

        for item in self._module_graphics_items.values():
            item.setFlag(QGraphicsItem.ItemIsMovable, not value)
            item.setLocked(value)

        for item in self._signal_graphics_items:
            item.setFlag(QGraphicsItem.ItemIsMovable, not value)

            # We use this value so lines in the signal can not be moved or edited
            item.is_locked = value

    @Slot()
    def save_layout(self) -> None:
        """
        Prompt the user for a file and save the scene layout to the file.

        This overwrites the previous content of the file.
        """
        # Prompt the user for a file
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        path = dialog.getSaveFileName()[0]

        # Open the given file erasing the previous content
        with open(path, "w") as fp:
            graphics_modules_data = {}
            ports_data = {}
            graphics_signals_data = {}

            graphics_modules = self.get_modules()
            for graphics_module in graphics_modules:
                pos = (graphics_module.x(), graphics_module.y())
                graphics_modules_data[graphics_module.getName()] = pos

                for port in graphics_module.getPorts():
                    visibility = port.isVisible()
                    orientation = int(port.getOrientation())
                    data = (port.x(), port.y(), orientation, visibility)
                    ports_data[port.getID()] = data

            for graphics_signal in self.get_signals():
                visibility = graphics_signal.isVisible()
                points = []
                for point in graphics_signal.getPoints():
                    points.append((point.x(), point.y()))
                data = (points, visibility)
                graphics_signals_data[graphics_signal.getID()] = data

            data = (graphics_modules_data, ports_data, graphics_signals_data)
            json.dump(data, fp)

        fp.close()

    @Slot()
    def load_default_layout(self) -> None:
        """
        Place all graphical modules in the scene according to the default
        layout, i.e. along a diagonal.
        """
        counter = 0
        for key in self._module_graphics_items:
            module_graphic = self._module_graphics_items[key]
            module_graphic.showPorts()
            counter += 1
            self.move_module_to(module_graphic, counter, counter)
        self.reset_signals()

    def load_layout_from_file(self, file_path: str) -> None:
        """
        Load a layout from a file.

        If at anypoint this function would error, the default layout
        is loaded instead.

        Parameters
        ----------
        file_path : str
            Path to a file containing a layout for the processor scene.
        """
        graphics_modules = self._module_graphics_items
        ports = {}
        graphics_signals = {}

        for graphics_module in graphics_modules.values():
            for port in graphics_module.getPorts():
                ports[port.getID()] = port

        for graphics_signal in self.get_signals():
            graphics_signals[graphics_signal.getID()] = graphics_signal

        # Open the file in 'read-only'
        with open(file_path, 'rb') as fp:
            data = json.load(fp)
            graphics_modules_data = data[0]
            ports_data = data[1]
            graphics_signals_data = data[2]

            for g_module_name, g_module_data in graphics_modules_data.items():
                g_module = graphics_modules[g_module_name]
                x = g_module_data[0]
                y = g_module_data[1]
                self._load_module(g_module, x, y)

            for port_id, port_data in ports_data.items():
                port = ports[int(port_id)]
                self._load_port(port, *port_data)

            for g_signal_id, g_signal_data in graphics_signals_data.items():
                g_signal = graphics_signals[int(g_signal_id)]
                self._load_signal(g_signal, *g_signal_data)

        fp.close()

    def _load_signal(
        self,
        signal: SignalGraphicsItem,
        signal_points: list[tuple[float, float]],
        visibility: bool,
    ) -> None:
        """
        Set the positions and visibility of a graphical signal. Helper method
        for loading a layout from file.

        Parameters
        ----------
        signal : SignalGraphicsItem
            Graphical signal to modify.
        signal_points : list[tuple[float, float]]
            List of points for visually drawing the signal as line.
        visibility : bool
            Visibility of the signal. ``True`` to show it, ``False`` to hide it.
        """
        qpoints = []
        # Turn points -> QPointF
        # list[int, int] -> QPointF object
        for point in signal_points:
            qpoints.append(QPointF(point[0], point[1]))

        # Set the new points
        signal.setPoints(qpoints)
        signal.setVisible(visibility)

    def _load_module(
        self,
        module: ModuleGraphicsItem,
        pos_x: float,
        pos_y: float,
    ) -> None:
        """
        Set the position of a graphical module in the scene. Helper method
        for loading a layout from file.

        Parameters
        ----------
        module : ModuleGraphicsItem
            Graphical module of which to set the position.
        pos_x : float
            Position on the x-axis in the scene.
        pos_y : float
            Position on the y-axis in the scene.
        """
        module.setX(pos_x)
        module.setY(pos_y)

    def _load_port(
        self,
        port: PortGraphicsItem,
        pos_x: float,
        pos_y: float,
        orientation: Orientation,
        visibility: bool,
    ) -> None:
        """
        Set position, orientation and visibility of a port of a graphical
        module. Helper method for loading a layout from file.

        Parameters
        ----------
        port : PortGraphicsItem
            Port to modify.
        pos_x : float
            Position on the x-axis in the scene.
        pos_y : float
            Position on the y-axis in the scene.
        orientation : Orientation
            Orientation of the port.
        visibility : bool
            Visibility of the port. ``True`` to show the port, ``False`` to
            hide it.
        """
        port.setOrientation(orientation)
        port.setX(pos_x)
        port.setY(pos_y)
        port.setVisible(visibility)
