import ast
import sys

from qtpy import QtCore, QtWidgets
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot, ws
from qtpy.QtWidgets import (
    QAction,
    QApplication,
    QErrorMessage,
    QMainWindow,
    QMessageBox,
)

from simudator.core.processor import Processor
from simudator.gui.breakpoint_window import BreakpointWindow
from simudator.gui.cpu_graphics_scene import CpuGraphicsScene
from simudator.gui.dialogs.lambda_breakpoint_dialog import LambdaBreakpointDialog
from simudator.gui.menu_bar import MainMenuBar
from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.pipeline import PipeLine
from simudator.gui.processor_handler import ProcessorHandler
from simudator.gui.simulation_toolbar import SimulationToolBar
from simudator.gui.view import View


class GUI(QMainWindow):
    """
    Main gui class for visualizing the SimuDator processor simulator.

    Graphics items for modules need to be created and added individually using
    the provided methods. Signals are also not added automatically but need
    not be created.

    Parameters
    ----------
    processor : Processor
        The simulated processor to visualize and control.
    """

    def __init__(self, processor: Processor):
        super().__init__()

        self._processor = processor
        self._processor_handler = ProcessorHandler(processor, parent=self)

        self.setWindowTitle("SimuDator")
        self._graphics_scene = CpuGraphicsScene()
        self._graphics_view = View(self._graphics_scene)
        # self.graphics_view.setDragMode(True)
        self._module_actions: dict[str, QAction] = {}

        self.setCentralWidget(self._graphics_view)

        self._error_msg_widget = QErrorMessage(self)

        self._create_menu_bar()
        self._create_tool_bar()

        self._pipeline = PipeLine()
        self.init_pipeline()

        # Set Style, THESE ARE TEST AND DONT WORK
        app = QApplication.instance()
        app.setStyleSheet("Qwidget.QMainWindow { background-color: yellow }")
        self.setAttribute(QtCore.Qt.WA_StyledBackground)

        # Add this so we can save the window in scope later, otherwise it disappears
        self._breakpoint_window = None

    def _create_menu_bar(self) -> None:
        """
        Create a main menu bar and connect its signals to appropriate slots of
        other widgets.
        """

        self._menu_bar = MainMenuBar(self)
        self.setMenuBar(self._menu_bar)

        # Connect signals for processor related actions
        self._menu_bar.load.connect(self._processor_handler.load_state)
        self._menu_bar.save.connect(self._processor_handler.save_state)
        self._menu_bar.reset.connect(self._processor_handler.reset)
        self._menu_bar.update_all_values.connect(
            self._processor_handler.toggle_value_update_on_run
        )
        self._menu_bar.set_delay.connect(self._processor_handler.set_update_delay)

        # Connect signals for managing the processor layout
        self._menu_bar.load_layout.connect(self._graphics_scene.load_layout)
        self._menu_bar.load_default_layout.connect(
            self._graphics_scene.load_default_layout
        )
        self._menu_bar.save_layout.connect(self._graphics_scene.save_layout)
        self._menu_bar.lock_layout.connect(self._graphics_scene.toggle_layout_lock)
        self._menu_bar.show_all_signals.connect(self._graphics_scene.show_all_signals)
        self._menu_bar.show_port_names.connect(self._graphics_scene.show_port_names)

        # Connect signal for managing breakpoints
        self._menu_bar.show_breakpoints.connect(self.openBreakpointWindow)

        self._processor_handler.running.connect(
            self._menu_bar.set_disabled_when_running
        )

    def _create_tool_bar(self) -> None:
        """
        Create a simulation toolbar and connect its signals to appropriate
        slots of other widgets.
        """
        self._toolbar = SimulationToolBar("Main toolbar", self)
        self.addToolBar(self._toolbar)
        self._toolbar.run.connect(self._processor_handler.run_simulation)
        self._toolbar.stop.connect(self._processor_handler.stop_simulation)
        self._toolbar.step.connect(self._processor_handler.step_cycles)
        self._toolbar.undo.connect(self._processor_handler.undo_cycles)
        self._toolbar.step_asm.connect(self._processor_handler.step_asm_instructions)
        self._toolbar.undo_asm.connect(self._processor_handler.undo_asm_instructions)
        self._processor_handler.cycle_changed.connect(self._toolbar.update_clock)
        self._processor_handler.running.connect(self._toolbar.set_disabled_when_running)

    def connectModuleActions(self, action_signals: []) -> None:
        """
        Connects modules with actions thorught ids (strings). Useful for breakpoints.

        This function takes a list of tuples with an
        action id (which is a predefined string) and a signal.
        It then connects the signal to a function depending on the action id.
        We do this so graphics items actions can access functionalty
        that is contolled by the gui and cpu. Mainly creating breakpoints.

        We do this instead of sending the graphics items the cpu or
        creating actions in the gpu. Sending the cpu directly to the
        graphics items could become problematic when many sources
        affect the cpu in unexpected ways, and creating these
        actions in the gpu would make the graphics items much less modular.
        """
        for action_id, signal in action_signals:
            match action_id:
                case "BP":
                    signal.connect(self.addStateBreakpoint)
                case "MEM_BP":
                    signal.connect(self.addMemoryBreakpoint)
                case "LMB_BP":
                    # LambdaBPs are not tested since we dont
                    # actually have any modules that make them yet
                    signal.connect(self.addLambdaBreakpoint)

    def init_pipeline(self) -> None:
        """Initialize the pipeline diagram.

        Sets its height, width and instructions.
        """
        size = self._processor.get_pipeline_dimensions()

        self._pipeline.set_height(size[0])
        self._pipeline.set_width(size[1])

        self._processor_handler.changed_instruction.connect(
            self._pipeline.set_instructions
        )
        # TODO: Find prettier way of making the processor handler emit its
        # signal for the current processor instructions
        self._processor_handler._signal_processor_changed()
        self._pipeline.show()

    def lambdaBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        lambdas = self._processor.get_breakpoint_lambdas()
        lambda_br_dialog = LambdaBreakpointDialog(lambdas, self)
        lambda_br_dialog.accepted.connect(self.addLambdaBreakpoint)

    def updateBreakpointWindow(self) -> None:
        """
        Updates the breakpoint window when new breakpoints are added.
        """
        # Don't do anything if window is closed
        if self._breakpoint_window is not None:
            self._breakpoint_window.update()

    """
    @Slot is used to explicitly mark a python method as a Qt slot
    and specify a C++ signature for it, which is used most commonly
    in order to select a particular overload.
    """

    @Slot(str, str, str)
    def editModuleState(self, module_name, state, value) -> None:
        """
        Tries to change the value of module to value given by user.

        Does this through the dialog opened by editModuleStateDialog.
        """
        try:
            parsed_value = ast.literal_eval(value)
        except SyntaxError as e:
            self._error_msg_widget.showMessage(str(e))
        else:
            module = self._processor.get_module(module_name)
            module_state = module.get_state()
            module_state[state] = parsed_value
            module.set_state(module_state)
            self._graphics_scene.update_modules()

    @Slot(str, str, str)
    def editMemoryContent(self, module_name: str, adress: str, value: str) -> None:
        """
        Tries to change the value at adress to value given by user.

        Does this through the dialog opened by editModuleStateDialog.
        """
        try:
            parsed_adress = int(adress, 16)
            parsed_value = ast.literal_eval(value)
        except SyntaxError as e:
            self._error_msg_widget.showMessage(str(e))
        except ValueError:
            self._error_msg_widget.showMessage(
                "You must enter a hexadecimal" "number preceeded by '0x' (e.g." "0xc3)."
            )
        else:
            module = self._processor.get_module(module_name)
            module_state = module.get_state()
            module_state['memory'][parsed_adress] = parsed_value
            module.set_state(module_state)
            self._graphics_scene.update_modules()

    @Slot(str, str, str)
    def addStateBreakpoint(self, module_name: str, state: str, value: str) -> None:
        """
        Tries to add a breakpoint to the module through the dialog opened by stateBreakpointDialog.
        """
        try:
            parsed_value = ast.literal_eval(value)
            self._processor.add_state_breakpoint(module_name, state, parsed_value)
            self.updateBreakpointWindow()
        except (ValueError, SyntaxError) as e:
            self._error_msg_widget.showMessage(str(e))

    @Slot(str, str)
    def addLambdaBreakpoint(self, lambda_name, kwargs_str) -> None:
        """
        Tries to add a breakpoint to the module through the dialog opened by lambdaBreakpointDialog.
        """
        try:
            lambda_kwargs = {}
            for kwarg in kwargs_str.split(' '):
                key, value = kwarg.split('=')
                lambda_kwargs[key] = ast.literal_eval(value)
            self._processor.add_lambda_breakpoint(lambda_name, **lambda_kwargs)
            self.updateBreakpointWindow()
        except (ValueError, SyntaxError) as e:
            self._error_msg_widget.showMessage(str(e))

    @Slot(str, str, str)
    def addMemoryBreakpoint(self, module_name: str, adress: str, value: str) -> None:
        """
        Tries to add a breakpoint to the module through the dialog opened by memoryBreakpointDialog.
        """
        try:
            parsed_adress = int(adress)
            parsed_value = int(value)
            self._processor.add_memory_breakpoint(
                module_name, parsed_adress, parsed_value
            )
            self.updateBreakpointWindow()
        except SyntaxError as e:
            self._error_msg_widget.showMessage(str(e))

    @Slot()
    def openBreakpointWindow(self) -> None:
        """
        Opens window for editing breakpoints.
        """
        self._breakpoint_window = BreakpointWindow(self._processor)
        self._breakpoint_window.show()

    def messageBox(self, message, title="Message") -> None:
        """
        Displays a simple box with the given message.
        """
        dlg = QMessageBox(self)
        dlg.setWindowTitle(title)
        dlg.setText(message)
        dlg.exec()

    def add_module_graphics_item(self, item: ModuleGraphicsItem) -> None:
        """
        Add a module graphics item to the graphics scene and connect its
        QT signals and slots.
        """
        self._graphics_scene.add_module(item)
        self.connectModuleActions(item.getActionSignals())
        self._processor_handler.changed.connect(item.update)

    def add_all_signals(self) -> None:
        """
        Add visual representations of all processor signals between all modules
        added to the GUI.
        """
        self._graphics_scene.add_all_signals()

    def load_layout(self, file_path: str) -> None:
        """Load a processor layout from file.

        Parameters
        ----------
        file_path : str
            Path to the file containing a processor layout to load.

        """
        self._graphics_scene.load_layout_from_file(file_path)


if __name__ == '__main__':
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    app = QApplication(sys.argv)
    gui = GUI(Processor())
    gui.show()
    app.exec()
