import typing

from PyQt5.QtWidgets import QAction
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QMenuBar, QWidget


class MainMenuBar(QMenuBar):
    """
    Main menu bar of the simudator GUI. Contains submenus for file I/O,
    layout managing and miscellaneous tools.

    Parameters
    ----------
    parent : QWidget | None
        Parent of the widget. The widget will be a new window if it is ``None``
        (default).
    """

    load = pyqtSignal()
    """
    QT signal emitted when the user has requested to load a processor state.
    """

    save = pyqtSignal()
    """
    QT signal emitted when the user has requested to save the processor state.
    """

    reset = pyqtSignal()
    """
    QT signal emitted when the user has request to reset the processor state.
    """

    load_layout = pyqtSignal()
    """
    QT signal emitted when the user has request to load a processor layout.
    """

    load_default_layout = pyqtSignal()
    """
    QT signal emitted when the user has request to load the default processor
    layout.
    """

    save_layout = pyqtSignal()
    """
    QT signal emitted when the user has request to save the processor layout.
    """

    lock_layout = pyqtSignal(bool)
    """
    QT signal emitted when the user has request to toggle the lock of editing
    the processor layout.
    """

    show_breakpoints = pyqtSignal()
    """
    QT signal emitted when the user has requested to show and manage
    the processor breakpoints.
    """

    show_all_signals = pyqtSignal(bool)
    """
    QT signal emitted when the user has requested to toggle showing all
    processor signals.
    """

    show_port_names = pyqtSignal(bool)
    """
    QT signal emitted when the user has requested to toggle showing the names
    of processor module ports.
    """

    update_all_values = pyqtSignal(bool)
    """
    QT signal emitted when the user has requested to see all values of
    processor modules update in real time when running the processor.
    """

    set_delay = pyqtSignal()
    """
    QT signal emitted when the user has requested to set the delay between
    clock cycles when running the processor.
    """

    def __init__(self, parent: typing.Optional[QWidget] = None) -> None:
        super().__init__(parent)

        # Create load action
        self._load_action = QAction("Load", self)
        self._load_action.setStatusTip("Load processor state from file")
        self._load_action.triggered.connect(self.load)

        # Create save action
        self._save_action = QAction("Save", self)
        self._save_action.setStatusTip("Save processor state to file")
        self._save_action.triggered.connect(self.save)

        self._reset_action = QAction("Reset", self)
        self._reset_action.setStatusTip("Reset processor")
        self._reset_action.triggered.connect(self.reset)

        # Create File menu for load and save
        file_menu = self.addMenu("&File")
        file_menu.addAction(self._load_action)
        file_menu.addAction(self._save_action)
        file_menu.addAction(self._reset_action)

        # create load layout action
        load_layout_action = QAction("Load layout", self)
        load_layout_action.setStatusTip("Loads the selected layout from file.")
        load_layout_action.triggered.connect(self.load_layout)

        # create load default layout action
        load_default_layout_action = QAction("Load default layout", self)
        load_default_layout_action.setStatusTip("Loads the default layout from file.")
        load_default_layout_action.triggered.connect(self.load_default_layout)

        # create save layout action
        save_layout_action = QAction("Save layout", self)
        save_layout_action.setStatusTip("Saves the current layout to a file.")
        save_layout_action.triggered.connect(self.save_layout)

        # create layout lock action
        self._lock_layout_action = QAction("Lock layout", self, checkable=True)
        self._lock_layout_action.setStatusTip("Lock layout so items can't be moved.")
        self._lock_layout_action.triggered.connect(self.lock_layout)

        # Create show signals actions
        self._signal_vis_action = QAction("Show signals", self, checkable=True)
        self._signal_vis_action.setChecked(True)
        self._signal_vis_action.setStatusTip("Toggle the visibility of signal.")
        self._signal_vis_action.triggered.connect(self.show_all_signals)

        # Create show port name actions
        self._port_vis_action = QAction("Show port names", self, checkable=True)
        self._port_vis_action.setChecked(True)
        self._port_vis_action.setStatusTip("Toggle the visibility of port names.")
        self._port_vis_action.triggered.connect(self.show_port_names)

        # Create Layout menu for layout actions
        layout_menu = self.addMenu("&Layout")
        layout_menu.addAction(load_layout_action)
        layout_menu.addAction(load_default_layout_action)
        layout_menu.addAction(save_layout_action)
        layout_menu.addAction(self._lock_layout_action)
        layout_menu.addAction(self._signal_vis_action)
        layout_menu.addAction(self._port_vis_action)

        # Create breakpoint window action
        self._breakpoint_action = QAction("Breakpoints", self)
        self._breakpoint_action.setStatusTip("Open breakpoint window.")
        self._breakpoint_action.triggered.connect(self.show_breakpoints)

        # Create 'update value' window button
        self._update_value_action = QAction(
            "Update values while running", self, checkable=True
        )
        self._update_value_action.setChecked(False)
        self._update_value_action.setStatusTip("Toggle value updates while running.")
        self._update_value_action.triggered.connect(self.update_all_values)

        # Create 'set delay' window button
        self._set_delay_action = QAction("Set update delay", self)
        self._set_delay_action.setStatusTip(
            "Sets the delay between each update when the cpu is running."
        )
        self._set_delay_action.triggered.connect(self.set_delay)

        # Create Tools menu for tool actions actions
        tools_menu = self.addMenu("&Tools")
        tools_menu.addAction(self._breakpoint_action)
        tools_menu.addAction(self._update_value_action)
        tools_menu.addAction(self._set_delay_action)

    @Slot(bool)
    def set_disabled_when_running(self, value: bool) -> None:
        """
        Toggle the menu bar between enabled and disabled. When disabled,
        actions which should not be performed while a processor is running
        are disabled and greyed out.

        Parameters
        ----------
        value : bool
            ``True`` to disable the menu bar, ``False`` to enable it.
        """
        self._load_action.setDisabled(value)
        self._save_action.setDisabled(value)
        self._reset_action.setDisabled(value)
