from enum import Enum

from qtpy.QtWidgets import QButtonGroup, QHBoxLayout, QHeaderView, QRadioButton

from simudator.gui.module_graphics_item.memory_graphic import (
    Memory,
    MemoryTable,
    MemoryWindow,
)


class Base(Enum):
    NONE = 0
    DECIMAL_SIGNED = 1
    DECIMAL_UNSIGNED = 2
    BINARY = 3
    HEXADECIMAL = 4


class ValueTooBig(Exception):
    """
    A class representing the error of a user inputting a value that does not fit in the memory.

    Used to differentiate between wrong input using the exception python raises by default (ValueError)
    and the error of a value that does not fit.
    """

    pass


class IntegerMemoryWindow(MemoryWindow):
    """
    A class showing the contents of a memory module in a new window.

    Parameters
    ----------
    memory_module: Memory
        An instance of the Memory base class.
    bit_length: int
        An integer specifying the number of bits of each address.
    """

    def __init__(self, memory_module: Memory, bit_length: int):

        # Do not let parent create edit/view buttons
        super().__init__(memory_module, False)

        # Remove the memory table the parent created
        self.layout.removeWidget(self._memory_table)

        # Add our own
        self._memory_table = IntegerMemoryTable(memory_module, bit_length)
        self.layout.addWidget(self._memory_table)

        # Create base buttons, they are exclusive by default
        # so need a seperate QButtonGroup since these four
        # have nothing to do with the edit/view buttons
        self.dec_signed_button = QRadioButton("Decimal Signed")
        self.dec_unsigned_button = QRadioButton("Decimal Unsigned")
        self.bin_button = QRadioButton("Binary")
        self.hex_button = QRadioButton("Hexadecimal")
        self.base_group = QButtonGroup()
        self.base_group.addButton(self.dec_signed_button, 1)
        self.base_group.addButton(self.dec_unsigned_button, 2)
        self.base_group.addButton(self.bin_button, 3)
        self.base_group.addButton(self.hex_button, 4)

        # Create edit/view buttons, they are exclusive by default
        self.edit_button = QRadioButton("Edit")
        self.view_button = QRadioButton("View")
        self.edit_group = QButtonGroup()
        self.edit_group.addButton(self.edit_button, 1)
        self.edit_group.addButton(self.view_button, 2)

        # Connect them to the 'set_base' function
        self.base_group.buttonClicked.connect(self._set_base)
        # Connect them to the 'set_edit' function
        self.edit_group.buttonClicked.connect(self._set_edit)

        # Create a layout that expands horizontally, so that
        # all buttons appear in one row
        self.first_button_layout = QHBoxLayout()
        self.first_button_layout.addWidget(self.dec_signed_button)
        self.first_button_layout.addWidget(self.dec_unsigned_button)
        self.first_button_layout.addWidget(self.bin_button)
        self.first_button_layout.addWidget(self.hex_button)
        self.layout.addLayout(self.first_button_layout)

        self.second_button_layout = QHBoxLayout()
        self.second_button_layout.addWidget(self.edit_button)
        self.second_button_layout.addWidget(self.view_button)
        self.layout.addLayout(self.second_button_layout)

        # Set the default base to decimal and disable editing
        # from start
        self.bin_button.toggle()
        self.view_button.toggle()
        self._set_base()
        self._set_edit()

        # Sets the size of each column to the smallest possible
        # width that allows all content in each box to be visible
        self._memory_table.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeToContents
        )

    def _set_base(self):
        pressed_button_id = self.base_group.checkedId()
        if pressed_button_id == 1:
            base = Base.DECIMAL_SIGNED
        elif pressed_button_id == 2:
            base = Base.DECIMAL_UNSIGNED
        elif pressed_button_id == 3:
            base = Base.BINARY
        else:
            base = Base.HEXADECIMAL

        self._memory_table.set_base(base)
        self._memory_table.update()


class IntegerMemoryTable(MemoryTable):
    """
    A class showing the contents of a memory module in a QTableWidget.

    This class assumes that the size of the memory module will remain constant.

    Parameters
    ----------
    memory_module: Memory
        An instance of the Memory base class.
    bit_length: int
        An integer specifying the number of bits of each address.
    column_size: int
        An integer specifying the number of columns, optional.

    """

    def __init__(self, memory_module: Memory, bit_length: int, column_size=-1):
        self._base = Base.DECIMAL_UNSIGNED
        self._bit_length = bit_length
        super().__init__(memory_module, column_size)

        self.update()
        # Signal used to detect when the user has edited a cell
        # This code can only be called after the initial self.update().
        # When the QTableWidget is initialized it currently fills each cell with empty strings.
        # It also signals cellChanged for each cell it initializes. Thus _on_cell_changed
        # is called for each cell and tries to save the cells empty string into the modules memory as
        # an integer.
        self.cellChanged.connect(self._on_cell_changed)

    def update(self):
        """
        Update the content of this widget to reflect the content of the memory module.
        """
        memory_content = self._memory.get_state()["memory"]
        for i in range(self._memory_size):
            value = memory_content[i]
            row = i // self._column_size
            col = i % self._column_size
            if self._base == Base.DECIMAL_UNSIGNED or self._base == Base.DECIMAL_SIGNED:
                value = self._get_dec_from_sign(value)
            elif self._base == Base.BINARY:
                value = str(bin(int(value)))[2:]  # remove '0b'
            else:  # Hex
                value = str(hex(int(value)))[2:]  # remove '0x'
            self.set_item(row, col, value)

    def set_base(self, base: Base):
        """
        Change what base the content is viewed in.

        Parameters
        ----------
        base: Base
            An instance of the base enum representing what base to change to.
        """
        self._base = base

    def _get_dec_from_sign(self, value: int) -> str:
        if self._base == Base.DECIMAL_UNSIGNED:
            return str(int(value))

        # Any value above or equal to 'threshold' must have the MSB set to 1
        threshold = 2 ** (self._bit_length - 1)

        if value < threshold:
            return str(int(value))

        # If the MSB is set to one subtract its value twice,
        # once so we dont count the positive value from it,
        # and the second time so we count it as negative
        return str(value - (threshold * 2))

    def _on_cell_changed(self, row, col):
        state = self._memory.get_state()

        item = self.item(row, col)
        value = item.text()
        max_value = 2**self._bit_length
        min_value = -(2 ** (self._bit_length - 1))

        # Turn every value into a positive int
        # in base 10
        # This makes converting between bases
        # much easier since we can assume what base
        # and format each value is in

        try:
            if self._base == Base.BINARY:
                value = int(value, 2)
            elif self._base == Base.HEXADECIMAL:
                value = int(value, 16)
            elif (
                self._base == Base.DECIMAL_SIGNED or self._base == Base.DECIMAL_UNSIGNED
            ):
                value = int(value)

            if value > max_value or value < min_value:
                raise ValueTooBig

        except ValueError:
            msg = None
            if self._base == Base.BINARY:
                msg = "You must enter a binary number (e.g. 0b101)."
            elif (
                self._base == Base.DECIMAL_SIGNED or self._base == Base.DECIMAL_UNSIGNED
            ):
                msg = "You must enter a decimal number (e.g. 107)."
            elif self._base == Base.HEXADECIMAL:
                msg = "You must enter a hexadecimal number (e.g. 0xc3)."
            self._errorMessageWidget.showMessage(msg)
            self.update()
            return

        except ValueTooBig:
            self._errorMessageWidget.showMessage(
                f"Input value does not fit within the bit length of {self._bit_length}."
            )
            self.update()
            return

        # Turn negative signed numbers into their positive unsigned
        # counter part
        # Ex 4 bits:
        # -7 = 1001
        # -7 + 16 % 16 = 9 = 1001 = -7
        value = (value + max_value) % max_value

        index = row * 4 + col
        state['memory'][index] = value
        self._memory.set_state(state)
