from math import ceil

from qtpy.QtCore import Qt
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import (
    QAction,
    QButtonGroup,
    QErrorMessage,
    QGraphicsRectItem,
    QGraphicsSimpleTextItem,
    QHBoxLayout,
    QHeaderView,
    QRadioButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from simudator.core.modules import Memory
from simudator.gui.dialogs.memory_content_dialog import MemoryContentDialog
from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem


class MemoryWindow(QWidget):
    """
    A class showing the contents of a memory module in a new window.

    Parameters
    ----------
    memory_module: Memory
        An instance of the Memory base class.
    edit_buttons: bool
        A bool deciding whether or not to add buttons to the widget
        for toggling between setting the memory contents to 
        editable or not.
    """

    def __init__(self, memory_module: Memory, edit_buttons=True):
        super().__init__()

        # Create a layout that expands vertically
        # so that the buttons can be displayed below the
        # memory content
        self.layout = QVBoxLayout(self)
        self._memory_table = MemoryTable(memory_module)
        self.layout.addWidget(self._memory_table)

        if edit_buttons:
            # Create edit/view buttons, they are exclusive by default
            self.edit_button = QRadioButton("Edit")
            self.view_button = QRadioButton("View")
            self.edit_group = QButtonGroup()
            self.edit_group.addButton(self.edit_button, 1)
            self.edit_group.addButton(self.view_button, 2)

            # Connect them to the 'set_base' function
            self.edit_group.buttonClicked.connect(self._set_edit)
            self.second_button_layout = QHBoxLayout()
            self.second_button_layout.addWidget(self.edit_button)
            self.second_button_layout.addWidget(self.view_button)
            self.layout.addLayout(self.second_button_layout)

            # Set the default base to decimal and disable editing
            # from start
            self.view_button.toggle()
            self._set_edit()

        # Sets the size of each column to the smallest possible
        # width that allows all content in each box to be visible
        self._memory_table.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeToContents
        )
        self._memory_table.update()

    def _set_edit(self):
        pressed_button_id = self.edit_group.checkedId()
        if pressed_button_id == 1:
            self._memory_table.set_editable(True)
        if pressed_button_id == 2:
            self._memory_table.set_editable(False)


class MemoryTable(QTableWidget):
    """
    A class showing the contents of a memory module in a QTableWidget.

    This class assumes that the size of the memory module will remain constant.

    Parameters
    ----------
    memory_module: Memory
        An instance of the Memory base class.
    column_size: int
        An integer specifying the number of columns, optional.

    """

    def __init__(self, memory_module: Memory, column_size=-1):
        super().__init__()
        self._memory = memory_module
        self._column_size = column_size
        self._memory_size = len(self._memory.get_state()["memory"])
        self._set_column_size()
        self.setColumnCount(self._column_size)
        rows = ceil(self._memory_size / self._column_size)
        self.setRowCount(rows)
        self._init_table_items(self._column_size, rows)
        self.setHorizontalHeaderLabels(["+" + str(i) for i in range(4)])
        self.set_editable(False)
        self._errorMessageWidget = QErrorMessage()

        vertical_headers = []
        for i in range(0, self._memory_size, self._column_size):
            vertical_headers.append(str(hex(i)))

        self.setVerticalHeaderLabels(vertical_headers)

        self.update()

        # Signal used to detect when the user has edited a cell
        # This code can only be called after the initial self.update().
        # When the QTableWidget is initialized it currently fills each celll with empty strings.
        # It also signals cellChanged for each cell it initializes. Thus _on_cell_changed
        # is called for each cell and tries to save the cells empty string into the modules memory as
        # an integer.
        self.cellChanged.connect(self._on_cell_changed)

    def update(self):
        """
        Update the content of this widget to reflect the content of the memory module.
        """
        memory_content = self._memory.get_state()["memory"]
        for i in range(self._memory_size):
            value = memory_content[i]
            row = i // self._column_size
            col = i % self._column_size
            value = str(value)
            self.set_item(row, col, value)

    def set_item(self, row: int, col: int, text: str) -> None:
        """Set the text at specified table cell to the given text.

        Parameters
        ----------
        row: int
             The items row position in the table.
        col: int
             The items column position in the table.
        text: str
             The text to be displayed.
        """
        item = QTableWidgetItem(text)

        self.setItem(row, col, item)

    def set_editable(self, editable: bool) -> None:
        if editable:
            self.setEditTriggers(self.AllEditTriggers)
        else:
            self.setEditTriggers(self.NoEditTriggers)

    def _init_table_items(self, cols: int, rows: int) -> None:
        for col in range(cols):
            for row in range(rows):
                item = QTableWidgetItem()
                self.setItem(row, col, item)

    def _set_column_size(self) -> None:
        """
        Set the column size to a reasonable value if the size was not given to the constructor.

        This function assumes that the attributes `column_size` and `memory_size` are set before it is called.
        """

        if not self._column_size == -1:
            return

        if self._memory_size > 200:
            self._column_size = 4
            return

        if self._memory_size > 100:
            self._column_size = 2
            return

        self._column_size = 1
        return

    def _on_cell_changed(self, row, col):
        state = self._memory.get_state()
        item = self.item(row, col)
        value = item.text()
        if value == "":
            return
        index = row * 4 + col
        state['memory'][index] = value
        self._memory.set_state(state)


class MemoryGraphicsItem(ModuleGraphicsItem):
    """
    Graphics module for a Memory module.
    """

    new_memory_breakpoint_signal = pyqtSignal(str, str, str)

    def __init__(self, memory_module: Memory, width: int = 150, height: int = 200):
        self._width = width
        self._height = height
        super().__init__(memory_module)
        self.memory_window = None

    def draw_graphics_item(self) -> None:
        # Create base backgrond rect
        self.baserect = QGraphicsRectItem(0, 0, self._width, self._height, self)

        # Make ports
        self.input = PortGraphicsItem(
            self.module.get_input_signal(), Orientation.RIGHT, self
        )
        self.input.setPos(self._width, self._height / 8)

        self.output = PortGraphicsItem(
            self.module.get_output_signal(), Orientation.RIGHT, self
        )
        self.output.setPos(self._width, self._height / 4)

        self.address = PortGraphicsItem(
            self.module.get_address_signal(), Orientation.DOWN, self
        )
        self.address.setPos(self._width / 4, self._height)

        self.control = PortGraphicsItem(
            self.module.get_control_signal(), Orientation.DOWN, self
        )
        self.control.setPos(self._width * 3 / 4, self._height)

        self.ports = [self.input, self.output, self.address, self.control]

        # Create name lable
        name_text = QGraphicsSimpleTextItem(self.state["name"], self)
        name_text.setPos(self._width / 2 - len(self.state["name"]) * self._CHAR_LEN, 0)

        # Create address lable
        self.address_text = QGraphicsSimpleTextItem(
            "address: " + str(self.state["current_address"]), self
        )
        self.address_text.setPos(self._width / 20, self._height / 8)

    def update(self):
        """
        Also update memory window here.
        """
        self.state = self.module.get_state()
        self.address_text.setText("address: " + str(self.state["current_address"]))
        if self.memory_window is not None:
            self.memory_window.update()

    def mousePressEvent(self, event):
        """
        Show content window if module is right clicked
        """
        return
        # TODO: comment =/= what fuction does
        if event.button() == Qt.MouseButton.RightButton:
            self.memory_window = MemoryWindow(self.module)
            self.memory_window.show()
            super().mousePressEvent(event)

    def showMemoryContents(self) -> None:
        """
        Create and show a MemoryWindow that displays the contents of the module.
        """
        self.memory_window = MemoryWindow(self._module)
        self.memory_window.show()

    def shouldIgnoreAction(self, action: QAction) -> bool:
        """
        Returns True if the given action should not be used in the context menu
        of this graphics item.
        """
        if action.data() == "MEM_ONLY":
            return False
        return super().shouldIgnoreAction(action)

    def generateActions(self) -> []:
        # Make parent actions and then add memory specific actions
        super().generateActions()

        memory_br_action = QAction('Add memory breakpoint', self)
        memory_br_action.triggered.connect(self.memoryBreakpointDialog)
        self.actions.append(memory_br_action)

        show_content_action = QAction('Show memory content', self)
        show_content_action.triggered.connect(self.showMemoryContents)
        self.actions.append(show_content_action)

    def memoryBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        self.memory_br_dialog = MemoryContentDialog(self.module)
        self.memory_br_dialog.accepted.connect(self.memoryBreakpointAccepted)

    def editMemoryContentDialog(self) -> None:
        """
        Opens dialog window for user to edit memory.
        """
        self.memory_edit_dialog = MemoryContentDialog(self.module)
        self.memory_edit_dialog.accepted.connect(self.editMemoryAccepted)

    @Slot(str, str, str)
    def memoryBreakpointAccepted(
        self, module_name: str, address: str, value: str
    ) -> None:
        """
        Takes the info from a breakpoint dialog and sends it to the gui
        """
        try:
            self.new_memory_breakpoint_signal.emit(module_name, address, value)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))

    @Slot(str, str, str)
    def editMemoryAccepted(self, module_name: str, address: str, value: str) -> None:
        """
        Takes the info from a edit memory dialog
        and updates the memory content accordingly.
        Also asks the gui to update.
        """
        try:
            parsed_address = int(address)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        else:
            module_state = self.module.get_state()
            module_state['memory'][parsed_address] = value
            self.module.set_state(module_state)
            self.module_edited.emit()
            self.update()

    def getActionSignals(self) -> []:
        # Do parent signals and then add memory specific signals
        signals = super().getActionSignals()
        signals.append(("MEM_BP", self.new_memory_breakpoint_signal))
        return signals
