import ast

from qtpy import QtCore
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import (
    QAction,
    QErrorMessage,
    QGraphicsItem,
    QGraphicsObject,
    QGraphicsRectItem,
    QGraphicsSceneContextMenuEvent,
    QGraphicsSceneMouseEvent,
    QGraphicsSimpleTextItem,
    QMenu,
)

from simudator.core.module import Module
from simudator.gui.color_scheme import ColorScheme
from simudator.gui.dialogs.module_state_dialog import ModuleStateDialog
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem


class ModuleGraphicsItem(QGraphicsObject, QGraphicsItem):
    """
    This is a general GraphicsItem that in theory should work for any module.
    It is also used as a kind of interface for other module graphics items.
    Can not guarantee visual appeal or readabiliy.
    """

    _SIGNAL_MARGIN = 20
    _STATE_MEMBER_MARGIN = 12
    _CHAR_LEN = 3
    _CHAR_HEIGHT = 8
    _WIDTH_MARGIN = 12
    moved = pyqtSignal(name="moved")
    new_state_breakpoint_signal = pyqtSignal(str, str, str)
    update_graphics_signal = pyqtSignal()

    module_edited = pyqtSignal()

    def __init__(self, module: Module, name: str = None):
        super().__init__()

        # Save module for later updates
        self.module = module

        self.errorMessageWidget = QErrorMessage()

        # Use modules name if no name is given
        if name is None:
            self.name = self.module.name
        else:
            self.name = name

        # Get information about module
        # Get inputs/outputs is not always implemented so might not show all ports
        self.state = module.get_state()
        self.input_signals = module.get_input_signals()
        self.output_signals = module.get_output_signals()

        # Have list of ports to reference for parent
        self.ports = []

        # start by setting ui lock of
        self.isLocked = False

        # Need this so ports always have something as a bounding rect
        self.baserect = QGraphicsRectItem(0, 0, 0, 0, self)

        self.actions = []
        self.generateActions()

        # Do general draw
        self.draw_graphics_item()

        for port in self.ports:
            self.moved.connect(port.moved)

        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFiltersChildEvents(False)

        # Changes the cursor when the module is hovered
        self.setCursor(QCursor(QtCore.Qt.PointingHandCursor))

        # Draw module last, which means it is the first to recivie any
        # mouse event
        self.setZValue(2.0)

    def draw_graphics_item(self) -> None:
        """
        This is a general draw function that will draw something for any module.
        Widgets for specific modules will override this function.
        """

        # Calculate height according to number of signals and states
        in_signals = len(self.input_signals)
        out_signals = len(self.output_signals)
        states = len(self.module.get_gui_state())
        inputs_height = (in_signals + 1) * self._SIGNAL_MARGIN
        outputs_height = (out_signals + 1) * self._SIGNAL_MARGIN
        members_height = states * self._CHAR_HEIGHT + self._STATE_MEMBER_MARGIN

        # Calculate width according to state character length
        longest_line_len = self.module.get_longest_line_len()
        height = max(inputs_height, outputs_height, members_height)
        width = longest_line_len * self._CHAR_LEN * 2 + 2 * self._WIDTH_MARGIN

        # Shadow
        self.shadow = QGraphicsRectItem(0, 0, width, height, self)
        self.shadow.setBrush(ColorScheme.ModuleShadow)
        self.shadow.setPen(ColorScheme.ModuleShadow)
        self.shadow.setPos(self.pos() + QtCore.QPointF(5, 10))

        # Create base background rectangle
        self.baserect = QGraphicsRectItem(0, 0, width, height, self)
        self.baserect.setBrush(ColorScheme.ModuleBackground)

        # Add ports for input signals
        self.draw_ports()

        # Create name lable
        name_text = QGraphicsSimpleTextItem(self.state["name"], self)
        name_text.setPos(width / 2 - len(self.state["name"]) * self._CHAR_LEN, 0)
        name_text.setBrush(ColorScheme.ModuleText)

        # Create state lables
        self.state_lables = {}
        gui_states = self.module.get_gui_state()
        for state_tuple in enumerate(gui_states):
            if state_tuple[1] != "name":
                text = state_tuple[1] + ": " + str(gui_states[state_tuple[1]])
                state_text = QGraphicsSimpleTextItem(text, self)
                state_text.setPos(width / 20, state_tuple[0] * self._CHAR_HEIGHT)
                state_text.setBrush(ColorScheme.ModuleText)
                self.state_lables[state_tuple[1]] = state_text

    def draw_ports(self) -> None:
        """
        Instantiates and draws the ports of this module graphics item.
        This does not create signal graphics items.
        """
        width = self.baserect.rect().width()
        height = self.baserect.rect().height()
        for signal_tuple in enumerate(self.input_signals):
            port = PortGraphicsItem(signal_tuple[1], Orientation.LEFT, self)
            height_spacing = height / (len(self.input_signals))
            port.setPos(0, (signal_tuple[0] + 0.5) * height_spacing)

            self.ports.append(port)

        # Add ports for output signals
        for signal_tuple in enumerate(self.output_signals):
            port = PortGraphicsItem(signal_tuple[1], parent=self)
            height_spacing = height / (len(self.output_signals))
            port.setPos(width, (signal_tuple[0] + 0.5) * height_spacing)
            self.ports.append(port)

    def showPorts(self) -> None:
        """
        Makes all ports of this module graphics item visible.
        """
        for port in self.ports:
            port.setVisible(True)

    def stateBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        self.state_br_dialog = ModuleStateDialog(self.module)
        self.state_br_dialog.okSignal.connect(self.stateBreakpointAccepted)

    def editModuleStateDialog(self) -> None:
        """
        Opens dialog where user can edit the state of a module.
        """
        self.state_br_dialog = ModuleStateDialog(self.module)
        self.state_br_dialog.okSignal.connect(self.editModuleState)

    @Slot(str, str, str)
    def editModuleState(self, module_name, state, value) -> None:
        """
        This function will be run when the edit module actions
        dialog is okey'd by the user.
        It will take the info from the dialog and edit the module accordinly.
        """
        try:
            parsed_value = ast.literal_eval(value)

            # ast.literal_eval only raises an error the the
            # parsed value isn't a valid data type.
            # If the user edits GRX and enters a string "a"
            # it will crash

            if isinstance(parsed_value, list):
                for value in parsed_value:
                    if isinstance(value, str):
                        raise TypeError
            if isinstance(parsed_value, str):
                raise TypeError

        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        except ValueError as e:
            self.errorMessageWidget.showMessage(
                "Value Error. Unable to parse input. Make sure it is in the correct base."
            )
        except TypeError as e:
            self.errorMessageWidget.showMessage(
                "Type Error. Unable to parse input. Make sure it is the correct type."
            )
        else:
            module_state = self.module.get_state()
            module_state[state] = parsed_value
            self.module.set_state(module_state)
            # Since we have changed a value we send a signal to the gui to update
            self.module_edited.emit()
            self.update()

    @Slot()
    def update(self):
        """
        Update the visuals of the graphics item to match it's module.
        Based on the generel build function so more specified
        widets will have to override it.
        """
        self.state = self.module.get_state()
        gui_state = self.module.get_gui_state()

        for name, value in gui_state.items():
            if name != "name":
                lable = self.state_lables[name]
                lable.setText(name + ": " + str(value))

    def boundingRect(self):
        """
        Used to make sure mouse clicks work on graphicsitem
        """
        return self.childrenBoundingRect()

    def paint(self, *args):
        # Paint is a pure virtual fuction that needs to be implemented
        pass

    def get_port(self, name: str) -> PortGraphicsItem:
        """
        Return the port graphics item with the given name of this module graphics item.
        """
        # TODO: This can be done with a dict if signals have uniqe names
        for port in self.ports:
            if port.getName() == name:
                return port
        return None

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)
        self.moved.emit()

    def contextMenuEvent(self, event: 'QGraphicsSceneContextMenuEvent') -> None:
        """
        Provides a context menu for this module graphics item when requested by
        a context menu event.
        """
        menu = self.getContextMenu()
        menu.exec_(event.screenPos())

    def getContextMenu(self) -> QMenu:
        """
        Returns a QMenu instance of the context menu used when right clicking
        this module graphics item.
        """
        menu = QMenu()
        for action in self.actions:
            menu.addAction(action)
            action.setData(self.name)
        return menu

    def generateActions(self) -> None:
        """
        Generates the default actions avalible in the context menu ofthe item.
        All other modules will also have these but may also add more actions.
        """

        # Create action to make all ports visible
        self.show_ports_action = QAction("Show all ports")
        self.show_ports_action.triggered.connect(self.showPorts)
        self.show_ports_action.setEnabled(not self.isLocked)
        self.actions.append(self.show_ports_action)

        # Create action to add breakpoint i items module
        state_br_action = QAction('Add breakpoint', self)
        state_br_action.triggered.connect(self.stateBreakpointDialog)
        self.actions.append(state_br_action)

        # Create action to edit items module
        edit_state_action = QAction('Edit module state', self)
        edit_state_action.triggered.connect(self.editModuleStateDialog)
        self.actions.append(edit_state_action)

    def getPortNames(self) -> list[str]:
        """
        Return a list of the names of all ports of this module graphics item.
        """
        return [p.getName() for p in self.ports]

    def getPorts(self) -> list[PortGraphicsItem]:
        """
        Return a list of all port graphics items of this module graphics item.
        """
        return self.ports

    def shouldIgnoreAction(self, action: QAction) -> bool:
        """
        Returns True if the given action should not be used in the context menu
        of this graphics item.
        """
        return action.data()

    def setLocked(self, locked: bool) -> None:
        self.isLocked = locked
        self.show_ports_action.setEnabled(not self.isLocked)
        for port in self.ports:
            port.setLocked(locked)

    def load_state_from_str(self, state_str) -> None:
        state = state_str.split()

        if state[0] != "x:":
            raise SyntaxWarning("Unexpected syntax when loading state")

        x_pos = float(state[1])

        if state[2] != "y:":
            raise SyntaxWarning("Unexpected syntax when loading state")

        y_pos = float(state[3])

        self.setPos(x_pos, y_pos)

    def save_state_as_str(self) -> str:
        x_pos = self.x()
        y_pos = self.y()
        name = self.name
        return f"{name}:\nx: {x_pos} y: {y_pos}\n"

    def getName(self) -> str:
        return self.name

    def getActions(self) -> []:
        """
        Will return a list of module specific actions that
        the gui will make available it the context menu.
        """
        return self.actions

    def getActionSignals(self) -> []:
        """
        This function returns a list of tuples with action ids and signals.
        The gui will connect these signals to functions depending on the id.
        This is mostly used so items can ask the gui to create breakpoints
        or ask it to updat functions depending on the id.
        This is mostly used so items can ask the gui to create breakpoints
        or ask it to update.
        """
        signals = []
        signals.append(("BP", self.new_state_breakpoint_signal))
        signals.append(("UPDATE", self.update_graphics_signal))
        return signals

    @Slot(str, str, str)
    def stateBreakpointAccepted(self, module_name: str, state: str, value: str) -> None:
        """
        This function takes the info from a breakpoint dialog and sends it to the gui.
        Other items might replace this function to parse the data.
        """
        try:
            self.new_state_breakpoint_signal.emit(module_name, state, value)
        except (ValueError, SyntaxError) as e:
            self.errorMessageWidget.showMessage(str(e))
