from typing import Any

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QTableWidget, QTableWidgetItem


class PipeLine(QTableWidget):
    """
    A class showing the current CPU instructions in a seperate window.

    The PipeLine class displays the instruction given by the CPU. It has no
    knowledge about the CPU itself, it only displays what is has been told to display.
    """

    def __init__(self):
        super().__init__()
        self.setEditTriggers(self.NoEditTriggers)

    @Slot(list)
    def set_instructions(self, instructions: list[tuple[str, int, int]]) -> None:
        """
        Give the pipeline the current CPU instructions.

        Parameters
        ----------
        instructions: A list of tuples of a string and two ints.
                      The string containing the current instruction where the two ints
                      represents the instructions col and row position.

        See Also
        --------
        simudator.core.processor.get_current_instructions :
            Instructions are set from get_current_instructions() in the Processor class.
        """

        self.instructions = instructions
        for i in range(len(self.instructions)):
            instruction = self.instructions[i][0]
            col = self.instructions[i][1]
            row = self.instructions[i][2]
            self.set_item(row, col, instruction)

    def set_height(self, height: int) -> None:
        """
        Sets the height of the pipeline.
        """
        self.setColumnCount(height)

    def set_width(self, width: int) -> None:
        """
        Sets the width of the pipeline.
        """
        self.setRowCount(width)

    def set_row_labels(self, labels: list[str]) -> None:
        """
        Sets the lable for each row. Does not gurantee that number of labels match rows.

        Depending on the CPU architecture the labels for each row
        can change format. In a pipelined CPU we want to assign a clock cycle
        to each instruction. In an ARM cpu we want the current
        CPU clock cycle for the first instruciton, and something else for the
        micro instrucitons.
        """
        raise NotImplemented

    def set_col_labels(self, labels: list[str]) -> None:
        """
        Sets the lable for each col. Does not gurantee that number of labels match cols.

        Depending on the CPU architecture the labels for each col
        can change format. In a pipelined CPU there is no reason to differentiate
        the different instrucitons, while in an ARM architecture there is.
        """
        raise NotImplemented

    def set_item(self, row: int, col: int, text: str) -> None:
        """
        Sets the text at position col row to the given text.
        """
        item = QTableWidgetItem(text)

        self.setItem(row, col, item)
