from qtpy import QtCore
from qtpy.QtCore import QObject, QThreadPool
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QErrorMessage, QFileDialog, QInputDialog, QMessageBox

from simudator.core.processor import Processor
from simudator.gui.run_continuously_thread import RunThread


class ProcessorHandler(QObject):
    """
    Wrapper class for interacting with the processor in the GUI.

    Parameters
    ----------
    processor : Processor
    parent : QObject | None
        Optional parent QObject that takes ownership of this QObject.
    """

    running = pyqtSignal(bool)
    """
    PyQT signal emitted when the processor has started or finished running. 
    Emits ``True`` when it has started and ``False`` when it has finished.
    """
    cycle_changed = pyqtSignal(int)
    """
    PyQT signal emitted when the current clock cycle count of the processor 
    has changed. Emits the current clock cycle count as an ``int``.
    """
    changed = pyqtSignal()
    """
    PyQT signal emitted when the processor has changed state.
    """
    changed_instruction = pyqtSignal(list)
    """
    PyQT signal emitted when the processor has changed the current instruction(s)
    being executed. Emits a list of instructions.

    See Also
    --------
    simudator.core.processor.get_current_instructions :
        Method of the processor used to get instructions that are emitted.
    """

    def __init__(
        self,
        processor: Processor,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._processor = processor
        # Threadpool for running cpu and gui concurrently
        self._threadpool = QThreadPool.globalInstance()

        # Used to set if all values in the gui should be updated each tick
        # or only the clock counter
        self._update_all_values = False

        # Used to set the update delay
        # Useful when watching the values being updated while running
        self._update_delay = 0.00

        # Used to lock some actions in ui when cpu is running in another thread
        # Using the cpu's internal status directly could case problems
        self._processor_running = False

        self._error_msg_box = QErrorMessage()

    def _signal_processor_changed(self) -> None:
        """
        Emit signals for signifying that the processor has changed its state.
        """
        self.cycle_changed.emit(self._processor.get_clock())
        self.changed.emit()
        self.changed_instruction.emit(self._processor.get_current_instructions())

    @Slot(int)
    def step_cycles(self, cycles: int):
        """
        Run some number of clock cycles of the processor.

        Parameters
        ----------
        Number of clock cycles to perform.
        """

        # Don't do steps if cpu is running
        if self._processor_running:
            return

        self._processor_running = True
        self.running.emit(True)
        simulation_thread = RunThread(
            self._processor,
            self._update_delay,
            False,
            False,
            cycles,
        )
        simulation_thread.processor_tick.connect(self.on_processor_tick)
        self._threadpool.start(simulation_thread)

    @Slot(int)
    def step_asm_instructions(self, instructions: int):
        """
        Run some numer of asm instructions.

        Parameters
        ----------
        instructions : int
            Number of asm instructions to perform.
        """

        # Don't do steps if cpu is running
        if self._processor_running:
            return

        self._processor_running = True
        self.running.emit(True)
        simulation_thread = RunThread(
            self._processor, self._update_delay, False, True, instructions
        )
        simulation_thread.processor_tick.connect(self.on_processor_tick)
        self._threadpool.start(simulation_thread)

    @Slot()
    def run_simulation(self) -> None:
        """
        Run the processor continuously until it is stopped by user input,
        reaches a breakpoint or halts on its own.
        """

        # Don't run if already running
        if self._processor_running:
            return

        # Create own thread for cpu simulation so gui dosent freeze
        self._processor_running = True
        self.running.emit(True)
        simulation_thread = RunThread(self._processor, self._update_delay)
        simulation_thread.processor_tick.connect(self.on_processor_tick)
        self._threadpool.start(simulation_thread)

    @Slot(int)
    def on_processor_tick(self, steps: int) -> None:
        """
        Called from other thread after every cpu tick.
        Will inform the user and update visuals.
        """

        # Update cpu clock counter every tick
        self.cycle_changed.emit(self._processor.get_clock())

        if self._update_all_values:
            self._signal_processor_changed()

        # A signal of 0 steps signifies end of execution, i.e. the CPU has
        # halted or run the specified amount of ticks
        # => Enable the relevant parts of the GUI again
        if steps == 0:
            self._processor_running = False
            self.running.emit(False)
            self._signal_processor_changed()

            # Inform user of reached break point
            if self._processor.breakpoint_reached:
                QMessageBox.information(
                    self.parent(),
                    "SimuDator",
                    f"Reached breakpoint: {str(self._processor.last_breakpoint)}",
                )

            # Inform user of halt
            if self._processor.should_halt():
                QMessageBox.information(
                    self.parent(), "SimuDator", "The processor halted."
                )

    @Slot()
    def stop_simulation(self) -> None:
        """
        Stop the processor.
        """
        self._processor.stop()

    def get_save_path(self) -> str:
        """
        Prompt the user for a file path meant for saving to file. Creates the
        specified file if it does not exist.
        """
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)

        return dialog.getSaveFileName()[0]

    def get_file_path(self) -> str:
        """
        Prompt the user for a file path meant for loading the processor state.
        """
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)

        return dialog.getOpenFileName()[0]

    @Slot()
    def load_state(self) -> None:
        """
        Load the processor state from selected file.
        """

        # not safe to load while cpu is running
        if self._processor_running:
            return

        path = self.get_file_path()
        try:
            self._processor.load_state_from_file(path)
        except KeyError:
            self._error_msg_box.showMessage("Could not load selected file.", "load_err")
        except ValueError:
            self._error_msg_box.showMessage("Selected file was empty.", "load_empty")
        except FileNotFoundError:
            # This error is triggered when no file was selected by
            # by the user. The user should know they did not select a
            # file and therefore there no need to display a box telling
            # them that.
            # self.messageBox("No valid file was selected.")
            pass
        else:
            QMessageBox.information(None, "SimuDator", "Loaded file successfully.")
            self._signal_processor_changed()

    @Slot()
    def reset(self) -> None:
        """
        Reset the processor to initial values and signal the processor state
        change.
        """

        # not safe to reset while cpu running
        if self._processor_running:
            return

        answer = QMessageBox.question(
            self.parent(),
            "Reset Processor",
            "Are you sure you want to reset the processor?",
        )

        if answer == QMessageBox.Yes:
            self._processor.reset()
            self._signal_processor_changed()

    @Slot()
    def save_state(self) -> None:
        """
        Save the state of the processor to file.

        This erases the previous content of the file before saving the data.
        """

        # Not safe to save while running
        if self._processor_running:
            return

        path = self.get_save_path()

        # Change file to the selected file if a file was selected
        if path == '':
            return

        res = self._processor.save_state_to_file(path)

        if res:
            QMessageBox.information(self.parent(), "SimuDator", "File saved.")
        else:
            self._error_msg_box.showMessage("Unable to save.", "save_err")

    @Slot(int)
    def undo_cycles(self, cycles: int) -> None:
        """
        Undo some number of processor clock cycles.

        Parameters
        ----------
        cycles : int
            Number of clock cycles to undo.
        """
        try:
            self._processor.load_cycle(max(self._processor.get_clock() - cycles, 0))
            self._signal_processor_changed()
        except (ValueError, IndexError):
            self._error_msg_box.showMessage("Unable to undo the cycle.")

    @Slot(int)
    def undo_asm_instructions(self, instructions: int) -> None:
        """
        Undo some number of asm instructions.

        Parameters
        ----------
        instructions : int
            Number of instructions to undo.
        """
        try:
            self._processor.undo_asm_instruction(instructions)
            self._signal_processor_changed()
        except (ValueError, IndexError):
            self._error_msg_box("Unable to undo the instruction.")

    @Slot()
    def toggle_value_update_on_run(self):
        """
        Toggle whether all values in the GUI related to the processor should be
        updated after each clock cycle or after having run several cycles at a
        time / running continuously. The clock cycle counter is always updated
        after each tick regardless.
        """
        self._update_all_values = not self._update_all_values

    @Slot()
    def set_update_delay(self):
        """
        Set the delay waited between each consecutive clock cycle when running
        several cycles or continuously. Prompts the user for the delay.
        """
        delay, ok = QInputDialog.getDouble(
            None, "Input Dialog", "Enter a float value:", decimals=5
        )
        if ok:
            self._update_delay = delay
