import time

from qtpy.QtCore import QObject, QRunnable
from qtpy.QtCore import Signal as pyqtSignal


class SignalHolder(QObject):
    """
    Helper class for RunThread as the thread cannot inherit from QObject to
    have signals directly. Instead, an instance of this class is used to hold
    signals for the thread.
    """

    processor_tick = pyqtSignal(int)


class RunThread(QRunnable):
    """
    This class is used to run the simulated cpu several ticks or continuously
    on a separate thread. This allows the user to interact with the GUI while
    the simulation is running.

    After each CPU tick, this thread will emit to its QT signal so that the GUI
    can update itself and possibly inform the user of when the execution
    has halted.
    """

    def __init__(
        self, cpu, delay: float, run_continuously=True, step_asm=False, steps=0
    ):

        super().__init__()
        self.cpu = cpu
        self.run_continuously = run_continuously
        self.steps = steps
        self.delay = delay
        self.step_asm = step_asm
        self.signals = SignalHolder()
        self.processor_tick = self.signals.processor_tick

    def run(self):
        if self.step_asm:

            while self.steps > 0:
                self.cpu.do_tick()
                self.processor_tick.emit(1)

                # We only care about asm instructions
                if self.cpu.is_new_instruction():
                    self.steps -= 1

                time.sleep(self.delay)

                if self.cpu.is_stopped:
                    break

        else:
            if self.run_continuously:
                self.cpu.unstop()
                while not self.cpu.is_stopped:
                    self.cpu.do_tick()
                    self.processor_tick.emit(1)
                    time.sleep(self.delay)

            else:
                for _ in range(self.steps):
                    self.cpu.do_tick()
                    self.processor_tick.emit(1)
                    time.sleep(self.delay)

                    if self.cpu.is_stopped:
                        break

        # Signal end of execution as having run 0 ticks
        self.processor_tick.emit(0)
