from PyQt5.QtWidgets import QAction
from qtpy import QtGui
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QLabel, QSpinBox, QStyle, QToolBar, QWidget


class SimulationToolBar(QToolBar):
    """
    A toolbar for basic controls of a processor simulation. This includes
    running and stopping the processor, stepping and undoing clock cycles and
    stepping and undoing asm instructions.

    Parameters
    ----------
    title : str
        Title of the toolbar.
    parent : QWidget | None
        Parent of the widget. The widget will be a new window if it is ``None``
        (default).
    """

    run = pyqtSignal()
    """
    QT signal emitted when the user has requested to run the processor
    continuously.
    """

    stop = pyqtSignal()
    """
    QT signal emitted when the user has requested to stop the processor.
    """

    step = pyqtSignal(int)
    """
    QT signal emitted when the user has requested to step some clock cycles
    of the processor.
    """

    undo = pyqtSignal(int)
    """
    QT signal emitted when the user has requested to undo some clock cycles of
    the processor.
    """

    step_asm = pyqtSignal(int)
    """
    QT signal emitted when the user has requested to step some asm instructions
    of the processor.
    """

    undo_asm = pyqtSignal(int)
    """
    QT signal emitted when the user has requested to undo some asm instructions
    of the processor.
    """

    def __init__(self, title, parent: QWidget | None = None):
        super().__init__(title, parent)
        # Run continuously button
        arrow_icon = self.style().standardIcon(QStyle.SP_MediaPlay)
        self._run_action = QAction(arrow_icon, "Run", self)
        self._run_action.setStatusTip("Run until halt")
        self._run_action.triggered.connect(self.run.emit)
        self.addAction(self._run_action)

        # Stop button
        stop_icon = self.style().standardIcon(QStyle.SP_MediaStop)
        self._stop_action = QAction(stop_icon, "Stop", self)
        self._stop_action.setStatusTip("Stop running")
        self._stop_action.triggered.connect(self.stop.emit)
        self.addAction(self._stop_action)

        # Separator and spacing between basic controls and stepping clock cycles
        spacing = QWidget()
        spacing.setFixedWidth(10)
        self.addWidget(spacing)
        self.addSeparator()
        spacing = QWidget()
        spacing.setFixedWidth(10)
        self.addWidget(spacing)

        # Clock cycle counter
        self._clock_cycle_label = QLabel("Clock cycles: 0", self)
        self._clock_cycle_label.setMinimumWidth(150)
        self.addWidget(self._clock_cycle_label)

        # Undo clock cycle button
        backward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekBackward)
        self._undo_action = QAction(backward_arrow_icon, "Undo", self)
        self._undo_action.setStatusTip("Undo the last processor tick")
        self._undo_action.triggered.connect(self._signal_undo)
        self.addAction(self._undo_action)

        # Step clock cycle button
        forward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekForward)
        self._step_action = QAction(forward_arrow_icon, "Step", self)
        self._step_action.setStatusTip("Run one clock cycle")
        self._step_action.triggered.connect(self._signal_step)
        self.addAction(self._step_action)

        # Spin box for the number of clock cycles to step/undo at a time
        self._jump_value_box = QSpinBox()
        self._jump_value_box.setMinimum(999999)
        self._jump_value_box.setMinimum(1)
        self._jump_value_box.setValue(1)
        self.addWidget(self._jump_value_box)

        # Separator and spacing between steping clock cycles and
        # stepping asm instructions
        spacing = QWidget()
        spacing.setFixedWidth(10)
        self.addWidget(spacing)
        self.addSeparator()
        spacing = QWidget()
        spacing.setFixedWidth(10)
        self.addWidget(spacing)

        # Asm instruction counter
        self._asm_label = QLabel("Assembler instructions: ", self)
        self.addWidget(self._asm_label)

        # Undo asm instruction button
        backward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekBackward)
        self._undo_asm_action = QAction(backward_arrow_icon, "Undo Asm", self)
        self._undo_asm_action.setStatusTip("Undo the last assembler instruction")
        self._undo_asm_action.triggered.connect(self._signal_undo_asm)
        self.addAction(self._undo_asm_action)

        # Step asm instruction button
        forward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekForward)
        self._step_asm_action = QAction(forward_arrow_icon, "Step Asm", self)
        self._step_asm_action.setStatusTip("Run one assembler instruction")
        self._step_asm_action.triggered.connect(self._signal_step_asm)
        self.addAction(self._step_asm_action)

        # Spin box for the number of asm instructions to step/undo at a time
        self._asm_jump_value_box = QSpinBox()
        self._asm_jump_value_box.setMinimum(999999)
        self._asm_jump_value_box.setMinimum(1)
        self._asm_jump_value_box.setValue(1)
        self.addWidget(self._asm_jump_value_box)

    def contextMenuEvent(self, a0: QtGui.QContextMenuEvent | None) -> None:
        # Override to remove the functionality of hiding the toolbar as there
        # is currently no way of showing it again
        pass

    @Slot(int)
    def update_clock(self, clock_cycle: int) -> None:
        """
        Set the displayed number of the clock cycle counter.

        Parameters
        ----------
        clock_cycle : int
            Number of clock cycles to display.
        """
        self._clock_cycle_label.setText(f"Clock cycles: {clock_cycle}")

    @Slot(bool)
    def set_disabled_when_running(self, value: bool) -> None:
        """
        Disable relevant parts of the toolbar related to controlling a
        processor. Intended to be used to disable stepping functionality
        while the processor is running.

        All buttons except for the stop button are disabled. The disabled
        buttons are greyed out and cannot be interacted with while the toolbar
        is disabled.

        Parameters
        ----------
        value : bool
            ``True`` to disable the toolbar, ``False`` to enable it.
        """
        self._run_action.setDisabled(value)
        self._step_action.setDisabled(value)
        self._undo_action.setDisabled(value)
        self._step_asm_action.setDisabled(value)
        self._undo_asm_action.setDisabled(value)

    def _signal_step(self):
        self.step.emit(self._jump_value_box.value())

    def _signal_undo(self):
        self.undo.emit(self._jump_value_box.value())

    def _signal_step_asm(self):
        self.step_asm.emit(self._asm_jump_value_box.value())

    def _signal_undo_asm(self):
        self.undo_asm.emit(self._asm_jump_value_box.value())
