from qtpy import QtCore, QtWidgets
from qtpy.QtGui import QWheelEvent
from qtpy.QtWidgets import QGraphicsView


class View(QGraphicsView):
    """
    This class views all QGraphicsItems for the user. It takes the
    QGraphicsScene as input and inherits all funcitonality from the
    QGraphicsView and overrides the wheelEvent function.
    This allows the users to navigate the view with their trackpads
    and zoom in/out with their trackpads + ctrl.
    """

    def __init__(self, QGraphicsScene):
        super().__init__(QGraphicsScene)
        self.scene = QGraphicsScene

    def wheelEvent(self, event: QWheelEvent | None) -> None:
        """
        Default behaviour if ctrl is not pressed, otherwise zoom in/out.
        """
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            # Factor above 1 zooms in, below zooms out
            factor = 1.03
            if event.angleDelta().y() < 0:
                factor = 0.97

            # If event got triggered due to the x axis, do nothing
            if event.pixelDelta().x() != 0:
                return

            view_pos = event.globalPosition()
            scene_pos = self.mapToScene(int(view_pos.x()), int(view_pos.y()))

            self.centerOn(scene_pos)
            self.scale(factor, factor)

            old_mapToScene = self.mapToScene(int(view_pos.x()), int(view_pos.y()))
            new_mapToScene = self.mapToScene(self.viewport().rect().center())

            delta = old_mapToScene - new_mapToScene

            self.centerOn(scene_pos - delta)

        else:
            # Default behaviour
            super().wheelEvent(event)
