import math

from qtpy.QtCore import QPointF
from qtpy.QtGui import QPolygonF
from qtpy.QtWidgets import (
    QErrorMessage,
    QGraphicsLineItem,
    QGraphicsPolygonItem,
    QGraphicsRectItem,
    QGraphicsSimpleTextItem,
)

from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem
from simudator.processor.mia.gui.bus_graphic import BusGraphicsItem
from simudator.processor.mia.modules.mia_grx import GRX


class GrxGraphicsItem(ModuleGraphicsItem):
    """
    Graphics module for mia's Grx module.
    """

    RECT_HEIGHT = 40
    RECT_WIDTH_PER_CHAR = 8
    TEXT_HEIGHT_MARGIN = 10
    TEXT_WIDTH_MARGIN = 5
    REGISTER_MARGIN = 10
    LINE_LENGTH = 30
    MUX_WIDTH = 80
    REGISTER_LENGTH = 4

    def __init__(self, module: GRX):
        self.register_text_labels = []
        super().__init__(module)

    def draw_graphics_item(self):
        # The width of the register
        # will be changed to fit text in registers but we
        # need it here to get the position
        rect_width = 80

        # Draw each register
        for index, register_value in enumerate(self.module.registers):
            # Make text to display for register
            name = "Gr" + str(index)

            # Mia uses 4 hex numbers to represent the registers
            hex_length = self.REGISTER_LENGTH
            value_text = f"0x{register_value:0{hex_length}x}"
            full_text = name + ": " + value_text[2:]

            # Make text label with text
            text = QGraphicsSimpleTextItem(full_text, self)
            text.setPos(
                self.TEXT_WIDTH_MARGIN,
                self.TEXT_HEIGHT_MARGIN
                + index * (self.REGISTER_MARGIN + self.RECT_HEIGHT),
            )
            self.register_text_labels.append(text)

            rect_width = (
                len(full_text) * self.RECT_WIDTH_PER_CHAR + self.TEXT_WIDTH_MARGIN
            )

            # Make rect that fits text
            QGraphicsRectItem(
                # rect posisition
                0,
                index * (self.REGISTER_MARGIN + self.RECT_HEIGHT),
                # rect size
                rect_width,
                self.RECT_HEIGHT,
                self,
            )

            # Add line to mux part, is purely cosmetical
            QGraphicsLineItem(
                # set first point pos
                rect_width,
                index * (self.REGISTER_MARGIN + self.RECT_HEIGHT)
                + self.RECT_HEIGHT / 2,
                # set other point pos
                rect_width + self.LINE_LENGTH,
                index * (self.REGISTER_MARGIN + self.RECT_HEIGHT)
                + self.RECT_HEIGHT / 2,
                self,
            )

        # Make mux polygon, is also purely cosmetical
        mux_height = (
            len(self.register_text_labels) * self.RECT_HEIGHT
            + (len(self.register_text_labels) - 1) * self.REGISTER_MARGIN
        )

        mux = QGraphicsPolygonItem(
            QPolygonF(
                # top left corner
                [
                    QPointF(0, 0),
                    # bottom left corner
                    QPointF(0, mux_height),
                    # bottom right corner
                    QPointF(self.MUX_WIDTH, mux_height - self.RECT_HEIGHT / 2),
                    # top right corner
                    QPointF(self.MUX_WIDTH, self.RECT_HEIGHT / 2),
                ]
            ),
            self,
        )

        # Move mux into position
        mux.setPos(rect_width + self.LINE_LENGTH, 0)

        # Have a invisible rect for ports bounding box
        self.baserect = QGraphicsRectItem(0, 0, self.MUX_WIDTH, mux_height)
        self.baserect.setPos(rect_width + self.LINE_LENGTH, 0)
        self.baserect.setVisible(False)

        # Add mux label
        name = "MUX"
        mux_text = QGraphicsSimpleTextItem(name, self)
        mux_text.setPos(
            rect_width + self.MUX_WIDTH / 2 + len(name) * self.RECT_WIDTH_PER_CHAR / 2,
            mux_height / 2 - self._CHAR_HEIGHT,
        )

        # Add ports to and from bus
        from_bus_port = PortGraphicsItem(
            self.module.signals["in_input"], Orientation.RIGHT, self
        )
        from_bus_port.setPos(
            rect_width + self.LINE_LENGTH + self.MUX_WIDTH,
            # Use the buses port margins so the ports align nicely
            mux_height / 2 - BusGraphicsItem.PORT_MARGIN / 2,
        )

        to_bus_port = PortGraphicsItem(
            self.module.signals["out_content"], Orientation.RIGHT, self
        )
        to_bus_port.setPos(
            rect_width + self.LINE_LENGTH + self.MUX_WIDTH,
            # Use the buses port margins so the ports align nicely
            mux_height / 2 + BusGraphicsItem.PORT_MARGIN / 2,
        )

        self.ports.append(from_bus_port)
        self.ports.append(to_bus_port)

    def update(self):
        self.state = self.module.get_state()
        for index, register_value in enumerate(self.module.registers):
            text = self.register_text_labels[index]
            name = "Gr" + str(index)

            # Mia uses 4 hex numbers for each register
            hex_length = self.REGISTER_LENGTH

            value_text = f"0x{register_value:0{hex_length}x}"
            full_text = name + ": " + value_text[2:]

            text.setText(full_text)
