import ast

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QErrorMessage, QGraphicsRectItem, QGraphicsSimpleTextItem

from simudator.core.modules import Memory
from simudator.gui.color_scheme import ColorScheme as CS
from simudator.gui.module_graphics_item.integer_memory_graphic import (
    IntegerMemoryWindow,
)
from simudator.gui.module_graphics_item.memory_graphic import (
    MemoryGraphicsItem,
    MemoryWindow,
)
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem
from simudator.processor.mia.gui.mia_memory_content_dialog import MiaMemoryContentDialog


class MiaMemoryGraphicsItem(MemoryGraphicsItem):
    """
    Graphics module for a Memory module.
    """

    def __init__(self, memory_module: Memory, **kwargs):
        super().__init__(memory_module, **kwargs)
        self.memory_window = None
        self._errorMessageWidget = QErrorMessage()

    def draw_graphics_item(self) -> None:
        # Same as normal memory but no control signal

        # Create base backgrond rect
        self.baserect = QGraphicsRectItem(0, 0, self._width, self._height, self)
        self.baserect.setPen(CS.ModuleOutline)
        self.baserect.setBrush(CS.ModuleBackground)

        # Make ports
        self.input = PortGraphicsItem(
            self.module.signals["in_input"], Orientation.RIGHT, self
        )
        self.input.setPos(self._width, self._height / 8)

        self.output = PortGraphicsItem(
            self.module.signals["out_content"], Orientation.RIGHT, self
        )
        self.output.setPos(self._width, self._height / 4)

        self.address = PortGraphicsItem(
            self.module.signals["in_address"], Orientation.DOWN, self
        )
        self.address.setPos(self._width / 4, self._height)

        self.ports = [self.input, self.output, self.address]

        # Create name lable
        name_text = QGraphicsSimpleTextItem(self.state["name"], self)
        name_text.setPos(self._width / 2 - len(self.state["name"]) * self._CHAR_LEN, 0)

        # Create address lable
        self.address_text = QGraphicsSimpleTextItem(
            "address: " + str(self.state["current_address"]), self
        )
        self.address_text.setPos(self._width / 20, self._height / 8)

    def showMemoryContents(self) -> None:
        """
        Create and show a MemoryWindow that displays the contents
        of the memory module associated with this graphics item.
        """
        bit_length = 16
        self.memory_window = IntegerMemoryWindow(self.module, bit_length)
        self.memory_window.show()

    def memoryBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        self.memory_br_dialog = MiaMemoryContentDialog(self.module)
        self.memory_br_dialog.accepted.connect(self.memoryBreakpointAccepted)

    @Slot(str, str, str)
    def memoryBreakpointAccepted(
        self, module_name: str, address: str, value: str
    ) -> None:
        """
        Same as prent function but also prases data so it is hexadecimal.
        """
        try:
            parsed_address = int(address, 16)
            parsed_value = ast.literal_eval(value)
            self.new_memory_breakpoint_signal.emit(
                module_name, str(parsed_address), str(parsed_value)
            )
        except SyntaxError as e:
            self._errorMessageWidget.showMessage(str(e))
        except ValueError:
            self._errorMessageWidget.showMessage(
                "You must enter a hexadecimal" "number preceeded by '0x' (e.g." "0xc3)."
            )

    @Slot(str, str, str)
    def editMemoryAccepted(self, module_name: str, address: str, value: str) -> None:
        """
        Same as prent function but also prases data so it is hexadecimal.
        """
        try:
            parsed_adress = int(address, 16)
            parsed_value = ast.literal_eval(value)
        except SyntaxError as e:
            self._errorMessageWidget.showMessage(str(e))
        except ValueError:
            self._errorMessageWidget.showMessage(
                "You must enter a hexadecimal" "number preceeded by '0x' (e.g." "0xc3)."
            )
        else:
            module_state = self.module.get_state()
            try:
                if parsed_adress < 0:
                    raise IndexError
                module_state['memory'][parsed_adress] = parsed_value
            except IndexError:
                self._errorMessageWidget.showMessage(
                    "Address entered was outside the memory space. Adress must be between 0 and 0xff."
                )
            else:
                self.module.set_state(module_state)
                self.update_graphics_signal.emit()
            module_state['memory'][parsed_adress] = parsed_value
            self.module.set_state(module_state)
            self.update_graphics_signal.emit()
