import sys
from typing import Any

from simudator.cli.cli import CLI
from simudator.core.modules import Flag
from simudator.core.modules.register import IntegerRegister
from simudator.core.processor import Processor, Signal
from simudator.gui.gui import GUI
from simudator.processor.mia.gui import (
    AluGraphicsItem,
    ArGraphicsItem,
    AsrGraphicsItem,
    BusGraphicsItem,
    FlagGraphicsItem,
    GrxGraphicsItem,
    HrGraphicsItem,
    IrGraphicsItem,
    MiaMemoryGraphicsItem,
    MicroMemoryGraphicsItem,
    PcGraphicsItem,
    SupcGraphicsItem,
    uPcGraphicsItem,
)
from simudator.processor.mia.modules.alu import ALU
from simudator.processor.mia.modules.ar import AR
from simudator.processor.mia.modules.asr import ASR
from simudator.processor.mia.modules.bus import Bus
from simudator.processor.mia.modules.hr import HR
from simudator.processor.mia.modules.ir import IR
from simudator.processor.mia.modules.lc import LC
from simudator.processor.mia.modules.mia_grx import GRX
from simudator.processor.mia.modules.mia_memory import MiaMemory
from simudator.processor.mia.modules.micro_memory import MicroMemory
from simudator.processor.mia.modules.micro_pc import MicroPC
from simudator.processor.mia.modules.pc import PC


class MIA_CPU(Processor):
    def __init__(self) -> None:
        super().__init__()
        # Creating all signals
        # Signals follow the naming convention 'to_from'

        # Add state variable names to skip when pretty printing the MIA
        # processor in the CLI
        self._ignore_keys += [
            "increment",
            "read_from_bus",
            "read_from_uADR",
            "decrement_by_one",
            "bus_id",
        ]

        # A signal used by all modules connected to the bus
        bus_control = Signal(self, "bus_control", (0, 0))

        # Dummy values used when the values in the constructor is not needed
        # unused = Signal(self)
        always_write_id = 999

        # Needed since optional arguments need to be in the correct order
        default_value = 0

        pm_bus = Signal(self, "PM-Bus", 0)
        bus_pm = Signal(self, "Bus-PM", 0)
        asr_bus = Signal(self, "asr_bus", 0)
        pm_asr = Signal(self, "pm_asr", 0)
        alu_ar = Signal(self, "alu_ar", 0)
        ar_alu = Signal(self, "ar_alu", 0)
        bus_ar = Signal(self, "bus_ar", 0)
        alu_hr = Signal(self, "alu_hr", 0)
        hr_alu = Signal(self, "hr_alu", 0)
        bus_hr = Signal(self, "bus_hr", 0)
        hr_bus = Signal(self, "hr_bus", 0)
        alu_bus = Signal(self, "alu_bus", 0)
        alu_uM = Signal(self, "alu_uM", 0)
        z_alu = Signal(self, "z_alu", 0)
        n_alu = Signal(self, "n_alu", 0)
        o_alu = Signal(self, "o_alu", 0)
        c_alu = Signal(self, "c_alu", 0)
        grx_bus = Signal(self, "grx_bus", 0)
        bus_grx = Signal(self, "bus_grx", 0)
        grx_control = Signal(self, "grx_control", 0)  # used to index registers with GRx
        m_control = Signal(self, "m_control", 0)  # used to index registers with M
        s_control = Signal(self, "s_control", 0)  # decides if GRx or M is used
        pc_uM = Signal(self, "pc_uM", 0)
        bus_pc = Signal(self, "bus_pc", 0)
        pc_bus = Signal(self, "pc_bus", 0)
        pm_bus = Signal(self, "pm_bus", 0)
        bus_pm = Signal(self, "bus_pm", 0)
        always_write_signal = Signal(self, value=(999, 0))
        uPC_k1 = Signal(self, "uPC_k1", 0)
        k1_ir = Signal(self, "k1_ir", 0)
        uPC_k2 = Signal(self, "uPC_k2", 0)
        ir_bus = Signal(self, "ir_bus", 0)
        bus_ir = Signal(self, "bus_ir", 0)
        SuPC_uPC = Signal(self, "SuPC_uPC", 0)
        uPC_SuPC = Signal(self, "uPC_SuPC", 0)
        uM_uPC = Signal(self, "uM_uPC", 0)
        uPC_uM_cont = Signal(self, "uPC_uM_cont", 0)
        uPC_uM = Signal(self, "uPC_uM", 0)
        bus_uM = Signal(self, "bus_uM", 0)
        lc_uM = Signal(self, "lc_uM", 0)
        lc_uM_uADR = Signal(self, "lc_uM_uADR", 0)
        uM_z = Signal(self, "uM_z", 0)
        uM_n = Signal(self, "uM_n", 0)
        uM_c = Signal(self, "uM_c", 0)
        uM_o = Signal(self, "uM_o", 0)
        uM_l = Signal(self, "uM_l", 0)
        lc_bus = Signal(self, "lc_bus", 0)
        l_lc = Signal(self, "l_lc", 0)

        # ASR specific
        asr_bus_id = 0b111
        asr_bit_length = 8
        asr = ASR(asr_bus, pm_asr, bus_control, asr_bus_id, asr_bit_length)

        # AR specific
        bus_ar_id = 0b100
        ar_bit_length = 16
        ar = AR(ar_alu, alu_ar, bus_ar, bus_control, bus_ar_id, ar_bit_length)

        # HR specific
        hr_bus_id = 0b101
        hr_bit_length = 16
        hr = HR(alu_hr, hr_alu, hr_bus, bus_hr, bus_control, hr_bus_id, hr_bit_length)

        # ALU specific
        alu = ALU(
            alu_ar,
            alu_bus,
            alu_hr,
            alu_uM,
            ar_alu,
            hr_alu,
            z_alu,
            n_alu,
            o_alu,
            c_alu,
            "ALU",
        )

        # PC specific
        pc_bus_id = 0b011
        pc = PC(pc_uM, pc_bus, bus_pc, bus_control, pc_bus_id)

        # PM specific
        pm_size = 256
        pm_bus_id = 0b010
        pm = MiaMemory(
            pm_bus,
            bus_pm,
            pm_asr,
            bus_control,
            pm_size,
            pm_bus_id,
            name="PM",
        )

        # GRX specific
        grx_bus_id = 0b110
        grx = GRX(
            bus_grx, grx_bus, bus_control, grx_control, m_control, s_control, grx_bus_id
        )

        # K1 specific
        k1_size = 16
        k1 = MiaMemory(
            always_write_signal,
            uPC_k1,
            k1_ir,
            always_write_signal,
            k1_size,
            always_write_id,
            name="K1",
        )

        # K2 specific
        k2_size = 4
        k2 = MiaMemory(
            always_write_signal,
            uPC_k2,
            m_control,
            always_write_signal,
            k2_size,
            always_write_id,
            name="K2",
        )

        # IR specific
        ir_bus_id = 0b001
        ir = IR(
            bus_ir,
            ir_bus,
            k1_ir,
            grx_control,
            m_control,
            bus_control,
            default_value,
            ir_bus_id,
        )

        # uPC specific
        uPC = MicroPC(uPC_uM_cont, uPC_k1, uPC_k2, SuPC_uPC, uPC_SuPC, uM_uPC, uPC_uM)

        # Su_PC specific
        SuPC_bit_length = 7
        SuPC_name = "SuPC"
        Su_PC = IntegerRegister(
            SuPC_uPC, uPC_SuPC, SuPC_bit_length, default_value, SuPC_name
        )

        # uM specific
        uM = MicroMemory(
            alu_uM,
            bus_control,
            bus_uM,
            s_control,
            pc_uM,
            lc_uM,
            lc_uM_uADR,
            uM_uPC,
            uPC_uM_cont,
            uPC_uM,
            uM_z,
            uM_n,
            uM_c,
            uM_o,
            uM_l,
        )

        # LC specific
        lc = LC(lc_uM, lc_bus, l_lc, lc_uM_uADR)

        # Flags
        z_flag = Flag(z_alu, uM_z, name="Z-Flag")
        n_flag = Flag(n_alu, uM_n, name="N-Flag")
        c_flag = Flag(c_alu, uM_c, name="C-Flag")
        o_flag = Flag(o_alu, uM_o, name="O-Flag")
        l_flag = Flag(l_lc, uM_l, name="L-Flag")

        # Bus
        bus = Bus(
            [bus_ir, bus_pm, bus_pc, bus_hr, bus_grx, bus_ar, bus_uM],
            [asr_bus, ir_bus, pm_bus, pc_bus, hr_bus, grx_bus, alu_bus, lc_bus],
        )

        ALLTHEMODULES = [
            uPC,
            lc,
            uM,
            Su_PC,
            ir,
            k2,
            k1,
            pc,
            ar,
            hr,
            grx,
            z_flag,
            n_flag,
            o_flag,
            l_flag,
            c_flag,
            pm,
            alu,
            asr,
            bus,
        ]

        for module in ALLTHEMODULES:
            self.add_module(module)

        self._micro_memory = uM

        self._lambdas = {}

    def is_new_instruction(self) -> bool:
        return self.get_module("uPC").get_state()["value"] == 0

    def get_current_instructions(self) -> list[tuple[str, int, int]]:
        asm_instr = self.get_asm_instruction()
        micro_instr = self.get_micro_instruction()
        asm = (asm_instr, 1, 0)
        micro = (micro_instr, 1, 1)
        asm_title = ("Assembler", 0, 0)
        micro_title = ("Micro", 0, 1)
        return [asm, micro, asm_title, micro_title]

    def get_micro_instruction(self) -> str:
        """
        Return a string containing the address for the current micro instruction in hex, e.g. 0a4b.
        """
        uM = self.get_module("uM").get_state()
        curr_instr = uM["curr_instr"]
        memory = uM["memory"]

        return '0x' + format(memory[curr_instr], '07x')

    def get_asm_instruction(self) -> str:
        """
        Return a string containing the label for the current assembly instruction.

        If the label doesnt exist, the string is empty.
        """
        ir_state = self.get_module("IR").get_state()
        op_code = ir_state["op"]
        return self.get_module("K1").get_label(int(op_code))

    def get_pipeline_dimensions(self) -> tuple[int, int]:
        """
        Mia should show the current assembly instruction and the current micro instruction.

        One Extra column to specify what the values represent.
        """
        return (2, 2)

    def should_halt(self) -> bool:
        micro_memory_state = self._micro_memory.get_state()
        return micro_memory_state["halt"]

    def launch_gui(self):
        from qtpy import QtCore
        from qtpy.QtWidgets import QApplication

        QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
        app = QApplication(sys.argv)
        self.load_state_from_file("mia_uppg3.txt")
        self.reset()
        gui = GUI(self)

        asr = self.get_module("ASR")
        ir = self.get_module("IR")
        pm = self.get_module("PM")
        pc = self.get_module("PC")
        alu = self.get_module("ALU")
        ar = self.get_module("AR")
        hr = self.get_module("HR")
        grx = self.get_module("GRx")

        bus_signal_pairs = [
            (asr.signals["in_content"], None),
            (ir.signals["in_input"], ir.signals["out_output"]),
            (pm.signals["in_input"], pm.signals["out_content"]),
            (None, None),
            (pc.signals["in_input"], pc.signals["out_content"]),
            (None, None),
            (alu.signals["in_input_bus"], None),
            (None, None),
            (None, ar.signals["out_content"]),
            (None, None),
            (hr.signals["in_content"], hr.signals["out_content"]),
            (None, None),
            (grx.signals["in_input"], grx.signals["out_content"]),
        ]

        gui.add_module_graphics_item(
            BusGraphicsItem(self.get_module("Bus"), bus_signal_pairs)
        )
        gui.add_module_graphics_item(uPcGraphicsItem(self.get_module("uPC")))
        gui.add_module_graphics_item(SupcGraphicsItem(self.get_module("SuPC")))
        gui.add_module_graphics_item(PcGraphicsItem(self.get_module("PC")))
        gui.add_module_graphics_item(ArGraphicsItem(self.get_module("AR")))
        gui.add_module_graphics_item(AsrGraphicsItem(self.get_module("ASR")))
        gui.add_module_graphics_item(HrGraphicsItem(self.get_module("HR")))
        gui.add_module_graphics_item(MicroMemoryGraphicsItem(self.get_module("uM")))
        gui.add_module_graphics_item(AluGraphicsItem(self.get_module("ALU")))
        gui.add_module_graphics_item(GrxGraphicsItem(self.get_module("GRx")))
        gui.add_module_graphics_item(IrGraphicsItem(self.get_module("IR")))

        flag_modules = ["Z-Flag", "N-Flag", "C-Flag", "O-Flag", "L-Flag", "LC"]
        for name in flag_modules:
            module = self.get_module(name)
            widget = FlagGraphicsItem(module)
            gui.add_module_graphics_item(widget)

        memory_modules = ["PM", "K1", "K2"]
        for name in memory_modules:
            module = self.get_module(name)
            widget = MiaMemoryGraphicsItem(module)
            gui.add_module_graphics_item(widget)

        gui.add_all_signals()

        gui.show()
        gui.load_layout("mia_layout")
        app.exec()

    def launch_cli(self):
        mia_cpu = MIA_CPU()
        cli = CLI(mia_cpu)
        mia_cpu.reset()
        cli.run()


if __name__ == "__main__":
    # this is bad terminal code, but who cares
    cpu = MIA_CPU()
    flag = None
    try:
        flag = sys.argv[1]
    except IndexError:
        pass

    if flag in ("gui", "g", "-gui"):
        cpu.launch_gui()
    else:
        cpu.launch_cli()
