from simudator.core.module import Module
from simudator.core.signal import Signal
from simudator.processor.mia.modules.alu_arithmetics import (
    add_binary_numbers,
    arithmetic_right_shift,
    bin_to_int,
    binary_and,
    binary_or,
    bitwise_not,
    int_to_bin,
    left_rotate,
    logical_left_shift,
    logical_right_shift,
)


class ALU(Module):
    """
    Module used for arithmetic logic calculations in MIA.

    Parameters
    ----------
    input_ar : Signal
        Input from the AR register, used as the first operand in calculations.
    input_bus : Signal
        Input from the bus, used as the second operand in calculations.
    input_hr : Signal
        Input from the HR register, used in 32-bit calculations.
    control : Signal
        Control signal for choosing ALU operation.
    output_ar : Signal
        Output signal to the AR register onto which the operation result is
        written.
    output_hr : Signal
        Output signal to the HR register onto which the lower 16-bits of the
        result of a 32-bit calculation are written.
    z_flag : Signal
        Output for setting the z-flag after a calculation.
    n_flag : Signal
        Output for setting the n-flag after a calculation.
    o_flag : Signal
        Output for setting the o-flag after a calculation.
    c_flag : Signal
        Output for setting the c-flag after a calculation.
    name : Signal
        Name of the ALU.
    """

    __slots__ = ("_z_value", "_n_value", "_o_value", "_c_value", "_hr_value")

    WORD_LENGTH = 16

    def __init__(
        self,
        input_ar: Signal,
        input_bus: Signal,
        input_hr: Signal,
        control: Signal,
        output_ar: Signal,
        output_hr: Signal,
        z_flag: Signal,
        n_flag: Signal,
        o_flag: Signal,
        c_flag: Signal,
        name: str,
    ):

        signals = {
            "in_input_ar": input_ar,
            "in_input_bus": input_bus,
            "in_control": control,
            "in_hr": input_hr,
            "out_result": output_ar,
            "out_hr": output_hr,
            "out_flag_z": z_flag,
            "out_flag_n": n_flag,
            "out_flag_o": o_flag,
            "out_flag_c": c_flag,
        }
        super().__init__(signals, name)

        # flag values
        self._z_value = 0
        self._n_value = 0
        self._o_value = 0
        self._c_value = 0

        # internal values
        self._hr_value = None

    def update_register(self) -> None:
        """Do nothing.

        The alu has no internal register.
        """
        pass

    def output_register(self) -> None:
        """Do nothing.

        The alu has no internal register.
        """
        pass

    def save_state_to_file(self, file_path: str) -> bool:
        """Do nothing.

        The alu is state-less.
        """
        return True

    def update_logic(self) -> None:
        """
        Will calculate and operation depending on given control signal
        and update outputs with result.
        """

        value_ar = self.signals["in_input_ar"].get_value()
        value_bus = self.signals["in_input_bus"].get_value()
        value_hr = self.signals["in_hr"].get_value()
        control_value = self.signals["in_control"].get_value()

        # Mask control value
        control_value = control_value & 0b1111

        # Define output value
        bin_ar = int_to_bin(value_ar, self.WORD_LENGTH)
        bin_bus = int_to_bin(value_bus, self.WORD_LENGTH)
        update_z_n_flags = True
        update_o_flag = False
        bin_res = None
        carry = self._c_value

        # Check control value for action
        match control_value:
            case 0b0000:
                # Do nothing
                update_z_n_flags = False
                return

            case 0b0001:
                # set output to only b
                update_z_n_flags = False
                bin_res = bin_bus

            case 0b0010:
                # calculate bitwise not of b
                update_z_n_flags = False
                bin_res = bitwise_not(bin_bus)

            case 0b0011:
                # returns 0
                bin_res = int_to_bin(0, self.WORD_LENGTH)

            case 0b0100:
                # return sum of a and b
                update_o_flag = True
                bin_res, carry = add_binary_numbers(bin_ar, bin_bus)

            case 0b0101:
                # returns the difference of a and b
                update_o_flag = True
                bin_bus = bitwise_not(bin_bus)
                bin_res, carry = add_binary_numbers(bin_ar, bin_bus, 1)

            case 0b0110:
                # returns the bitwise AND of a and b
                bin_res = binary_and(bin_ar, bin_bus)

            case 0b0111:
                # returns the bitwise OR of a and b
                bin_res = binary_or(bin_ar, bin_bus)

            case 0b1000:
                # return sum of a and b without changing flags
                update_z_n_flags = False
                update_o_flag = False
                bin_res, _ = add_binary_numbers(bin_ar, bin_bus)

            case 0b1001:
                # returns the leftshifted AR
                bin_res, carry = logical_left_shift(bin_ar)

            case 0b1010:
                # returns the leftshifted ARHR
                long_word = bin_ar + int_to_bin(value_hr, self.WORD_LENGTH)
                bin_res, carry = logical_left_shift(long_word)
                self._hr_value = bin_to_int(bin_res[self.WORD_LENGTH :])

            case 0b1011:
                # returns the arithmetic righshifted AR
                bin_res, carry = arithmetic_right_shift(bin_ar)

            case 0b1100:
                # returns the arithmetic righshifted ARHR
                long_word = bin_ar + int_to_bin(value_hr, self.WORD_LENGTH)
                bin_res, carry = arithmetic_right_shift(long_word)
                self._hr_value = bin_to_int(bin_res[self.WORD_LENGTH :])

            case 0b1101:
                # returns the logically righshifted AR
                bin_res, carry = logical_right_shift(bin_ar)

            case 0b1110:
                # returns AR one step rotated left
                bin_res, carry = left_rotate(bin_ar)

            case 0b1111:
                # returns ARHR one step rotated left
                long_word = bin_ar + int_to_bin(value_hr, self.WORD_LENGTH)
                bin_res, carry = left_rotate(long_word)
                self._hr_value = bin_to_int(bin_res[self.WORD_LENGTH :])

            case _:
                return

        # Update output signal and flags
        self._update_flags(
            bin_ar,
            bin_bus,
            bin_res,
            carry,
            update_z_n_flags,
            update_o_flag,
        )

        if len(bin_res) == 2 * self.WORD_LENGTH:
            # In case of a 32-bit calculation, extract the upper 16-bits
            # that should go into the AR register
            # (Not done earlier to properly update the flags as a 32-bit number)
            bin_res = bin_res[: self.WORD_LENGTH]
        res = bin_to_int(bin_res)
        self.signals["out_result"].update_value(res)
        self.signals["out_flag_z"].update_value(self._z_value)
        self.signals["out_flag_n"].update_value(self._n_value)
        self.signals["out_flag_o"].update_value(self._o_value)
        self.signals["out_flag_c"].update_value(self._c_value)
        self.signals["out_hr"].update_value(self._hr_value)

    def print_module(self) -> None:
        print(
            "",
            self.name,
            "\n -----",
            "\n a_value: ",
            self.signals["in_input_ar"].get_value(),
            "\n b_value: ",
            self.signals["in_input_bus"].get_value(),
            "\n control_signal: ",
            self.signals["in_control"].get_value(),
        )

    def reset(self) -> None:
        self._z_value = 0
        self._n_value = 0
        self._o_value = 0
        self._c_value = 0
        self._hr_value = None

    def _update_flags(
        self,
        a: list[int],
        b: list[int],
        res: list[int],
        carry: int,
        z_n: bool = True,
        o: bool = False,
    ) -> None:
        """Update the outputted values to the flags of the ALU depending on
        the operands and resulting from the ALU operation.

        It is possible to not update specific flags. For convenience, the
        c-flag is always updated with the supplied carry value.

        Parameters
        ----------
        a : list[int]
            First ALU operand.
        b : list[int]
            Second ALU operand.
        res : list[int]
            Result of the ALU operation.
        carry : int
            Carry bit from the ALU operation.
        z_n : bool
            ``True`` if the z- and n-flags should be updated, ``False`` if not.
        o : bool
            ``True`` if the o-flag should be updated, ``False`` if not
        """

        if z_n:
            # Set the z flag if the result is 0
            if bin_to_int(res) == 0:
                self._z_value = 1
            else:
                self._z_value = 0

            # Set the n flag if the most significant bit is 1,
            # i.e. the number is negative if interpreted as two's complement
            if res[0] == 1:
                self._n_value = 1
            else:
                self._n_value = 0

        if o:
            # Set the o flag if the MSB is the same for both a and b and this
            # bit is different to the MSB of the result
            if (a[0] == b[0]) and (a[0] != res[0]):
                self._o_value = 1
            else:
                self._o_value = 0

        self._c_value = carry
