def bin_to_int(binary_number: list[int]) -> int:
    """
    Convert from binary bit list representation to integer.

    Parameters
    ----------
    binary_number : list[int]
        Binary number represented as a list of bits.

    Returns
    -------
    int
        The binary number converted to an integer.
    """
    number = 0
    word_length = len(binary_number)

    # Get values for bits and add them
    for bit_index in range(0, word_length):
        number += binary_number[bit_index] * 2 ** (word_length - bit_index - 1)

    return number


def int_to_bin(number: int, word_length: int) -> list[int]:
    """Convert a non-negative integer to a binary number represented as a list
    of bits.

    Parameters
    ----------
    number : int
        Non-negative integer to convert to binary.
    word_length : int
        Number of bits to represent the binary number.

    Returns
    -------
    list[int]
        List of bits representing the number converted to binary.
    """
    # Get list representation with correct length
    binary_number = [0 for _ in range(word_length)]

    # We can now look at the other bits
    for bit_index in range(0, word_length):
        # If number is larger then bit we subtract the
        # bits decimal value from the number and make the bit 1.
        if number >= 2 ** (word_length - bit_index - 1):
            number -= 2 ** (word_length - bit_index - 1)
            binary_number[bit_index] = 1

    return binary_number


def bin_to_int_twocomp(binary_number: list[int]) -> int:
    """Convert from two's complement binary bit list representation to integer.

    Parameters
    ----------
    binary_number : list[int]
        Binary number in two's complement, represented as a bit list.

    Returns
    -------
    int
        The binary number converted to an integer.
    """
    number = 0
    word_length = len(binary_number)

    # two's complement representation so most significant
    # bit will be negative
    if binary_number[0] == 1:
        number = -(2 ** (word_length - 1))

    # Get values for other bits
    for bit_index in range(1, word_length):
        number += binary_number[bit_index] * 2 ** (word_length - bit_index - 1)

    return number


def bitwise_not(binary_number: list[int]) -> list[int]:
    """Perform bitwise not on a binary number.

    Parameters
    ----------
    binary_number : list[int]
        Binary number represented as a list of bits.

    Returns
    -------
    list[int]
        Resulting bit-flipped binary number as a list of bits.
    """

    # Make empty binary number for result
    new_binary_number = [0 for _ in range(len(binary_number))]

    # Change each bits value
    for bit_index in range(len(binary_number)):
        if binary_number[bit_index] == 0:
            new_binary_number[bit_index] = 1
        elif binary_number[bit_index] == 1:
            new_binary_number[bit_index] = 0

    return new_binary_number


def add_binary_numbers(
    a: list[int], b: list[int], carry: int = 0
) -> tuple[list[int], int]:
    """Add two binary numbers together.

    Assumes both numbers are represented with the same number of bits,
    i.e. have the same word length.

    Parameters
    ----------
    a : list[int]
        First operand represented as a list of bits.
    b : list[int]
        Second operand represented as a list of bits.
    carry : int
        Carry in bit for the least significant bit. Useful for subtraction.

    Returns
    -------
    sum : list[int]
        Resulting binary sum represented as a list of bits.
    carry : int
        Carry bit from performing the summation.
    """

    # Make new binary number to store results
    word_length = len(a)
    binary_sum = [0 for _ in range(word_length)]

    # Need carry for calculation
    # carry = 0

    # Do addition bit by bit and save carry between each loop
    for bit_index in range(word_length - 1, -1, -1):
        bit_sum = a[bit_index] + b[bit_index] + carry
        match bit_sum:
            case 0:
                binary_sum[bit_index] = 0
                carry = 0
            case 1:
                binary_sum[bit_index] = 1
                carry = 0
            case 2:
                binary_sum[bit_index] = 0
                carry = 1
            case 3:
                binary_sum[bit_index] = 1
                carry = 1

    return (binary_sum, carry)


def binary_and(a: list[int], b: list[int]) -> list[int]:
    """
    Calculate bitwise and between two binary numbers.

    Assumes both numbers are represented with the same number of bits,
    i.e. have the same word length.

    Parameters
    ----------
    a : list[int]
        First operand represented as a list of bits.
    b : list[int]
        Second operand represented as a list of bits.

    Returns
    -------
        Resulting binary number as a list of bits.
    """
    word_length = len(a)
    binary_and = [0 for _ in range(word_length)]

    for binary_index in range(word_length):
        binary_and[binary_index] = a[binary_index] & b[binary_index]

    return binary_and


def binary_or(a: list[int], b: list[int]) -> list[int]:
    """Calculate bitwise or between two binary numbers

    Assumes both numbers are represented with the same number of bits,
    i.e. have the same word length.

    Parameters
    ----------
    a : list[int]
        First operand represented as a list of bits.
    b : list[int]
        Second operand represented as a list of bits.

    Returns
    -------
        Resulting binary number as a list of bits.
    """
    word_length = len(a)
    binary_or = [0 for _ in range(word_length)]

    for binary_index in range(word_length):
        binary_or[binary_index] = a[binary_index] | b[binary_index]

    return binary_or


def logical_left_shift(bin_num: list[int]) -> tuple[list[int], int]:
    """Logically left shift a binary number.

    Parameters
    ----------
    bin_num : list[int]
        Binary number represented as a list of bits.

    Returns
    -------
    shifted_bin_num : list[int]
        Resulting shifted binary number as a list of bits.
    carry : int
        Carry bit from the left shift, i.e. the bit that is shifted out.
    """
    new_bin_num = bin_num[1:] + [0]
    carry = bin_num[0]

    return new_bin_num, carry


def logical_right_shift(bin_num: list[int]) -> tuple[list[int], int]:
    """Logically right shift a binary number.

    Parameters
    ----------
    bin_num : list[int]
        Binary number represented as a list of bits.

    Returns
    -------
    shifted_bin_num : list[int]
        Resulting shifted binary number as a list of bits.
    carry : int
        Carry bit from the right shift, i.e. the bit that is shifted out.
    """
    new_bin_num = [0] + bin_num[:-1]
    carry = bin_num[-1]
    return new_bin_num, carry


def arithmetic_right_shift(bin_num: list[int]) -> tuple[list[int], int]:
    """Arithmetically right shift a binary number.

    Parameters
    ----------
    bin_num : list[int]
        Binary number represented as a list of bits.

    Returns
    -------
    shifted_bin_num : list[int]
        Resulting shifted binary number as a list of bits.
    carry : int
        Carry bit from the right shift, i.e. the bit that is shifted out.
    """
    new_bin_num = [bin_num[0]] + bin_num[:-1]
    carry = bin_num[-1]
    return new_bin_num, carry


def left_rotate(bin_num: list[int]) -> tuple[list[int], int]:
    """Rotate the bits of a number to the left.

    Parameters
    ----------
    bin_num : list[int]
        Binary number represented as a list of bits.

    Returns
    -------
    rotated_bin_num : list[int]
        Resulting rotated binary number as a list of bits.
    carry : int
        Carry bit from the rotation, i.e. the bit that is shifted "out".
    """
    new_bin_num = bin_num[1:] + [bin_num[0]]
    carry = bin_num[0]
    return new_bin_num, carry
