from simudator.core.modules.register import IntegerRegister
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class HR(IntegerRegister, MiaBusConnector):
    """
    Register for saving large AlU calculations in MIA.
    """

    # Python does not allow multiple inherintence if more than one of the
    # parent classes uses __slots__. Thus we also includes the __slots__
    # from MiaBusConnector.
    __slots__ = ("bus_id", "bus_control_s")

    def __init__(
        self,
        alu_input_signal: Signal,
        alu_output_signal: Signal,
        bus_input_signal: Signal,
        bus_output_signal: Signal,
        bus_control_signal: Signal,
        bus_id: int,
        bit_length: int,
        name="HR",
    ) -> None:
        IntegerRegister.__init__(
            self, bus_input_signal, bus_output_signal, bit_length, name=name
        )
        MiaBusConnector.__init__(self, bus_control_signal, bus_id)

        alu_input_signal.add_destination(self)
        self.signals["in_alu"] = alu_input_signal
        self.signals["out_alu"] = alu_output_signal

    def update_register(self) -> None:
        """
        Updates the internal values from the signals from the bus and
        the alu. It only updates the bus when asked to, and will always
        update the alu signal.
        """

        if self.read_from_bus():
            IntegerRegister.update_register(self)

        elif self.signals["in_alu"].get_value() is not None:
            input_value = self.signals["in_alu"].get_value()
            mask = 2**self._bit_length - 1
            self._value = input_value & mask

    def output_register(self) -> None:
        """
        Outputs the internal values to the signals to the bus and
        the alu. It only outputs the bus when asked to, and will always
        output the alu signal.
        """

        self.signals["out_alu"].update_value(self._value)

        if self.write_to_bus():
            IntegerRegister.output_register(self)

        else:
            self.signals["out_content"].update_value(None)

    def update_logic(self):
        self.output_register()

    def print_module(self):
        print(
            "",
            self.name,
            "\n -----",
            "\n value: ",
            hex(self._value),
            "\n bit length: ",
            self._bit_length,
        )
