from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class LC(Module):
    """A class representing the loop counter. It is controlled by the
    signal uM_control which determines if it should read from bus,
    increase by one or read from mM_uADR.
    """

    __slots__ = (
        "_value",
        "_read_from_bus",
        "_read_from_uADR",
        "_decrement_by_one",
        "_bit_length",
    )

    def __init__(
        self,
        mM_control: Signal,
        bus_input: Signal,
        l_flag: Signal,
        mM_uADR: Signal,
        name="LC",
        value=0,
        bit_length=8,
    ) -> None:
        """
        Parameters
        ----------
        mM_control : Signal
            A signal connection from the micro memory
            to the loop counter. This allows the micro memory to send bit
            12 and 13 to the loop counter. Bit 12 and 13 decides the
            loop counters behaviour.
        bus_input : Signal
            A signal connection from the bus to the loop
            counter. The loop counter reads from this signal when it reads
            from the bus.
        l_flag : Signal
            A signal connection from the loop counter to the
            L flag. The loop counter writes to this signal when it needs to
            update the L flag.
        mM_uADR : Signal
            A signal connection from the micro memory to
            the loop counter. This allows the loop counter to read the 7
            least significant bits from the uADR field.
        name : str
            Optional name of the loop counter.
        value : int
            Optional start value of the loop counter.
        bit_length : int
            Optional bit length of the loop counter.
        """

        # signals
        signals = {
            "in_control": mM_control,
            "in_input": bus_input,
            "in_address": mM_uADR,
            "out_flag_l": l_flag,
        }

        super().__init__(signals, name)

        # the value of the loop counter
        self._value = value

        # helper variables
        self._read_from_bus = False
        self._read_from_uADR = False
        self._decrement_by_one = False

        # bit length and mask
        self._bit_length = bit_length

    def update_register(self) -> None:
        """Reads bit 12 and 13 from the micro memory and updates the
        loop counter.
        0 - does nothing.
        1 - decrements the loop counter by one.
        2 - loads the 8 least significant bits from the bus.
        3 - loads the 7 least significant bits from the uADR field.
        """

        match self.signals["in_control"].get_value():
            case 0b00:  # LC is not effected
                self._decrement_by_one = False
                self._read_from_bus = False
                self._read_from_uADR = False

            case 0b01:  # Decrement by one
                self._decrement_by_one = True
                self._read_from_bus = False
                self._read_from_uADR = False

            case 0b10:  # Load 8 least significant bits from bus
                self._decrement_by_one = False
                self._read_from_bus = True
                self._read_from_uADR = False

            case 0b11:  # LCs 7 least significant bits are loaded from uADR
                self._decrement_by_one = False
                self._read_from_bus = False
                self._read_from_uADR = True

        if self._read_from_bus:
            input_value = self.signals["in_input"].get_value()
            mask = 2**self._bit_length - 1
            self._value = input_value & mask

        if self._read_from_uADR:
            input_value = self.signals["in_address"].get_value()
            mask = 2**self._bit_length - 1
            self._value = input_value & mask

        if self._decrement_by_one:
            self._value -= 1

            # underflow correctly
            if self._value < 0:
                mask = 2**self._bit_length - 1
                self._value = mask

    def output_register(self) -> None:
        """The loop counter will only output to the L flag, this is
        handled in 'update_logic'.
        """
        pass

    def update_logic(self) -> None:
        """When the loop counter reaches zero, set the l flag to 1.
        Otherwise set it to zero.
        """
        if self._value == 0:
            self.signals["out_flag_l"].update_value(1)
        else:
            self.signals["out_flag_l"].update_value(0)

    def get_state(self) -> dict[str, Any]:
        """Returns a dict of the loop counter state.
        These states are changable via set_states.

        Returns
        -------
        dict[Any]
            The state of the loop counter.
        """
        state = dict()
        state["name"] = self.name
        state["value"] = self._value
        state["read_from_bus"] = self._read_from_bus
        state["read_from_uADR"] = self._read_from_uADR
        state["decrement_by_one"] = self._decrement_by_one
        return state

    def get_parameter(self) -> dict[str, Any]:
        """Return a dictionary of the parameters of the loop counter,
        i.e. the bit length of the counter.

        Returns
        -------
        dict[str, Any]
            The parameter of the loop counter.
        """
        parameter = super().get_parameter()
        parameter["bit_length"] = self._bit_length
        return parameter

    def set_state(self, state: dict[str, Any]) -> None:
        """Sets the loop counter state to one given in dict."""
        self.name = state["name"]
        self._value = state["value"]
        if "bit_length" in state:
            self._bit_length = state["bit_length"]
        if "read_from_bus" in state:
            self._read_from_bus = state["read_from_bus"]
        if "read_from_uADR" in state:
            self._read_from_uADR = state["read_from_uADR"]
        if "decrement_by_one" in state:
            self._decrement_by_one = state["decrement_by_one"]

    def reset(self) -> None:
        """Resets the loop counter to 0."""
        self._value = 0
        self._read_from_bus = False
        self._read_from_uADR = False
        self._decrement_by_one = False

    def save_state_to_file(self, file_path: str) -> bool:
        content = self.name + ":\n"
        content += "value: " + hex(self._value)[2:] + "\n\n"
        return super()._helper_save_state_to_file(file_path, content)

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        self._value = int(string_pair[1], 16)

    def print_module(self) -> None:
        print(
            "",
            self.name,
            "\n -----",
            "\n value: ",
            hex(self._value),
            "\n decrement: ",
            self._decrement_by_one,
            "\n read from uADR: ",
            self._read_from_uADR,
            "\n read from bus: ",
            self._read_from_bus,
        )
