from simudator.core.module import Module
from simudator.core.signal import Signal


class MicroPC(Module):
    """
    A class representing the micro program counter (uPC) of the MIA-system.
    The module reads signal data from the memory modules K1 and K2, the
    register SuPC and the micro memory uM. The module writes signal
    data to SuPC and uM. A control signal is used to determine which signal
    to read from or write to.
    """

    __slots__ = ("_value", "_bit_length")

    def __init__(
        self,
        control_signal: Signal,
        from_k1: Signal,
        from_k2: Signal,
        to_supc: Signal,
        from_supc: Signal,
        to_um: Signal,
        from_um: Signal,
        name: str = "uPC",
    ) -> None:

        self._value = 0

        # Signals
        signals = {
            "in_control": control_signal,
            "in_k1": from_k1,
            "in_k2": from_k2,
            "in_supc": from_supc,
            "in_um": from_um,
            "out_supc": to_supc,
            "out_um": to_um,
        }

        super().__init__(signals, name)

    def update_register(self) -> None:
        match self.signals["in_control"].get_value():
            case 0b000:
                self._value += 1

                # Overflow
                if self._value == 128:
                    self._value = 0

            case 0b001:
                self._value = self.signals["in_k1"].get_value()
            case 0b010:
                self._value = self.signals["in_k2"].get_value()
            case 0b011:
                self._value = 0
            case 0b100:
                self._value = self.signals["in_um"].get_value()
            case 0b101:
                self._value = self.signals["in_supc"].get_value()
            case 0b110:
                self.signals["out_supc"].update_value(self._value + 1)
                self._value = self.signals["in_um"].get_value()

    def output_register(self) -> None:
        self.signals["out_um"].update_value(self._value)

    def get_state(self) -> dict:
        state = super().get_state()
        state["value"] = self._value
        return state

    def get_gui_state(self) -> dict:
        state = {
            "name": self.name,
            "value": self._value,
        }
        return state

    def set_state(self, state: dict) -> None:
        super().set_state(state)
        self._value = state["value"]

    def reset(self) -> None:
        """
        Resets the micro program counter to 0.
        """
        self._value = 0

    def update_logic(self) -> None:
        """
        The uPC has no logic.
        """
        pass

    def save_state_to_file(self, file_path: str) -> bool:
        content = self.name + ":\n"
        content += "value: " + hex(self._value)[2:] + "\n\n"
        return super()._helper_save_state_to_file(file_path, content)

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        self._value = int(string_pair[1], 16)

    def print_module(self) -> None:
        print("", self.name, "\n -----", "\n value: ", hex(self._value))
