from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class PC(Module, MiaBusConnector):
    """A program counter moduled after the mia lab.

    The module have two
    control signals, p and bus_control. If the p signal is true the
    program counter increases its value by one, if the bus_control
    is true (the id to read from bus is the pc's id) the pc takes the
    value on the bus. If both are true it does nothing.
    """

    # Python does not allow multiple inherintence if more than one of the
    # parent classes uses __slots__. Thus we also includes the __slots__
    # from MiaBusConnector.
    __slots__ = ("bus_id", "bus_control_s", "_value", "_increase_by_one", "_bit_length")

    def __init__(
        self,
        p: Signal,
        bus_input: Signal,
        bus_output: Signal,
        bus_control: Signal,
        bus_id: int,
        name="PC",
        value=0,
    ) -> None:
        MiaBusConnector.__init__(self, bus_control, bus_id)

        signals = {
            "in_control": p,
            "in_input": bus_input,
            "out_content": bus_output,
        }
        Module.__init__(self, signals, name)

        self._value = value
        self._increase_by_one = False
        self._bit_length = 8

    def output_register(self) -> None:
        """Output the value of the program counter to the bus."""
        if self.write_to_bus():
            self.signals["out_content"].update_value(self._value)
        else:
            self.signals["out_content"].update_value(None)

    def update_register(self) -> None:
        """Updates the value of the PC according to the signals sent by the micro memory.

        If the program counter recives 'True' from both signals it
        should do nothing and wait for the next cycle.
        """

        self._increase_by_one = self.signals["in_control"].get_value()

        if self._increase_by_one and self.read_from_bus():
            return

        mask = 2**self._bit_length - 1
        if self.read_from_bus():
            # Read the 8 lowest bits from the bus
            self._value = self.signals["in_input"].get_value() & mask

        if self._increase_by_one:
            self._value += 1

            # overflow correctly
            if self._value > mask:
                self._value = 0

    def update_logic(self) -> None:
        self.output_register()

    def get_state(self) -> dict[str, Any]:
        """Returns a dict of the program counter state.

        These states are changable via set_states.

        Returns
        -------
        dict[str, Any]
            The state of the program counter.
        """
        state = dict()
        state["name"] = self.name
        state["value"] = self._value
        state["increment"] = self._increase_by_one
        return state

    def get_parameter(self) -> dict[str, Any]:
        """Return a dictionary of the parameters of the program counter,
        i.e. the bit length of the counter.

        Returns
        -------
        dict[str, Any]
            The paramter of the program counter.
        """
        paramter = Module.get_parameter(self)
        paramter["bit_length"] = self._bit_length
        return paramter

    def get_gui_state(self) -> dict:
        state = {
            "name": self.name,
            "value": self._value,
        }
        return state

    def set_state(self, state: dict[str, Any]) -> None:
        """Sets the program counter state to one given in dict."""
        self.name = state["name"]
        self._value = state["value"]
        if "increment" in state:
            self._increase_by_one = state["increment"]

    def reset(self) -> None:
        """Reset the program counter to 0."""
        self._value = 0
        self._increase_by_one = False

    def save_state_to_file(self, file_path: str) -> bool:
        content = self.name + ":\n"
        content += "value: " + hex(self._value)[2:] + "\n\n"
        return super()._helper_save_state_to_file(file_path, content)

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        self._value = int(string_pair[1], 16)

    def print_module(self) -> None:
        print("", self.name, "\n -----", "\n value: ", hex(self._value))
