from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.alu import ALU

cpu = Processor()
input_signal_a = Signal(cpu)
input_signal_b = Signal(cpu)
input_signal_hr = Signal(cpu)
control_signal = Signal(cpu)
output_signal = Signal(cpu)
hr_signal = Signal(cpu)
z_signal = Signal(cpu, value=0)
n_signal = Signal(cpu, value=0)
o_signal = Signal(cpu, value=0)
c_signal = Signal(cpu, value=0)
alu = ALU(
    input_signal_a,
    input_signal_b,
    input_signal_hr,
    control_signal,
    output_signal,
    hr_signal,
    z_signal,
    n_signal,
    o_signal,
    c_signal,
    "alu",
)
cpu.add_module(alu)


def create_test_case(op_ar, op_bus, op_code, res, z=None, n=None, o=None, c=None):
    """
    Create a test case for an ALU operation, optionally checking the flags set.

    Parameters
    ----------
    op_ar : int
        Input value to the ALU from the AR register, i.e. operand A.
    op_bus : int
        Input value to the ALU from the bus, i.e. operand B.
    op_code : int
        ALU operation code.
    res: int
        Result from the ALU operation.
    z : int | None
        Expected z-flag value after the operation. Set it to ``None`` if the
        flag should not be affected by the operation.
    n : int | None
        Expected n-flag value after the operation. Set it to ``None`` if the
        flag should not be affected by the operation.
    o : int | None
        Expected o-flag value after the operation. Set it to ``None`` if the
        flag should not be affected by the operation.
    c : int | None
        Expected c-flag value after the operation. Set it to ``None`` if the
        flag should not be affected by the operation.

    Returns
    -------
    tuple
        Test case for the ALU.
    """
    return op_ar, op_bus, op_code, res, z, n, o, c, None, None


def create_hr_test_case(op_ar, op_hr, op_code, ar_res, hr_res, z, n, c):
    """
    Create a test case for a 32-bit ALU operation, i.e. one using the values
    of the AR and HR registers together.

    The 32-bit ALU operations always affect the z-, n- and c-flag while
    not affecting the o-flag. As such, those 3 flags must be provided with
    expected values.

    Parameters
    ----------
    op_ar : int
        Input value to the ALU from the AR register.
    op_hr : int
        Input value to the ALU from the HR register.
    op_code : int
        ALU operation code.
    ar_res: int
        Result to the AR register from the ALU operation.
    hr_res: int
        Result to the HR register from the ALU operation.
    z : int
        Expected z-flag value after the operation.
    n : int
        Expected n-flag value after the operation.
    c : int
        Expected c-flag value after the operation.

    Returns
    -------
    tuple
        Test case for the ALU.
    """
    return op_ar, 0, op_code, ar_res, z, n, None, c, op_hr, hr_res


def check_test_case(test_case):
    """
    Assert that the ALU passes a test case for some operation.

    Parameters
    ----------
    test_case : tuple
        Test case to test the ALU against.
    """
    (op_1, op_2, op_code, res, z, n, o, c, hr_in, hr_out) = test_case
    input_signal_a.update_value(op_1)
    input_signal_b.update_value(op_2)
    input_signal_hr.update_value(hr_in)
    control_signal.update_value(op_code)

    # Flags set to None in the test case should not be affected by the ALU
    # instruction, so save their values before performing the operation
    if z is None:
        z = alu._z_value
    if n is None:
        n = alu._n_value
    if o is None:
        o = alu._o_value
    if c is None:
        c = alu._c_value

    alu.update_logic()

    assert output_signal.get_value() == res
    assert hr_signal.get_value() == hr_out
    assert z_signal.get_value() == z
    assert n_signal.get_value() == n
    assert o_signal.get_value() == o
    assert c_signal.get_value() == c


def check_test_cases(test_cases):
    """
    Assert that the ALU passes some test cases for some operations.

    Note that the ALU is not reset between each test case, meaning that the
    order of the test cases may matter depending on how they affect the flags.

    Parameters
    ----------
    test_case : list[tuple]
        List of test cases to test the ALU against.
    """
    for case in test_cases:
        check_test_case(case)


def test_nop():
    op_code = 0b0000
    # Set the flag values of the ALU and make sure that they are not affected
    alu.reset()
    alu._z_value = 1
    alu._n_value = 1
    alu._o_value = 1
    alu._c_value = 1
    z_signal.set_value(1)
    n_signal.set_value(1)
    o_signal.set_value(1)
    c_signal.set_value(1)
    test_case = create_test_case(10, 5, op_code, None)
    check_test_case(test_case)


def test_read_bus():
    op_code = 0b0001
    # Set the flag values of the ALU and make sure that they are not affected
    alu.reset()
    alu._z_value = 1
    alu._n_value = 1
    alu._o_value = 1
    alu._c_value = 1
    test_cases = (
        create_test_case(0x0000, 0x0000, op_code, 0x0000),
        create_test_case(0x0000, 0x0174, op_code, 0x0174),
        create_test_case(0x0000, 0xFFFFF, op_code, 0xFFFF),  # Overflow
        create_test_case(0x0AEF, 0x0360, op_code, 0x0360),  # Random AR value
    )
    check_test_cases(test_cases)


def test_bitwise_not():
    op_code = 0b0010
    # Set the flag values of the ALU and make sure that they are not affected
    alu.reset()
    alu._z_value = 1
    alu._n_value = 1
    alu._o_value = 1
    alu._c_value = 1
    test_cases = (
        # Random values in the AR input which should not affect the result
        create_test_case(0x0000, 0x0000, op_code, 0xFFFF),
        create_test_case(0x0F38, 0x0004, op_code, 0xFFFB),
        create_test_case(0x0000, 0xFFFF, op_code, 0x0000),
        create_test_case(0x2E6D, 0x8531, op_code, 0x7ACE),
    )
    check_test_cases(test_cases)


def test_reset_ar():
    op_code = 0b0011
    # Set the o- and c-flag values of the ALU and make sure that they are
    # not affected
    alu.reset()
    alu._o_value = 1
    alu._c_value = 1
    # (Random values in the AR and bus input which should not affect the result)
    test_case = create_test_case(0xFD12, 0x34B7, op_code, 0, 1, 0)
    check_test_case(test_case)


def test_add():
    op_code = 0b0100
    test_cases = (
        create_test_case(0x0001, 0x0000, op_code, 0x0001, 0, 0, 0, 0),
        create_test_case(0x0000, 0x0001, op_code, 0x0001, 0, 0, 0, 0),
        # Test all possible flag combinations
        create_test_case(0x0004, 0x0008, op_code, 0x000C, 0, 0, 0, 0),
        create_test_case(0xC000, 0x4001, op_code, 0x0001, 0, 0, 0, 1),
        create_test_case(0x8000, 0x8001, op_code, 0x0001, 0, 0, 1, 1),
        create_test_case(0x8000, 0x0001, op_code, 0x8001, 0, 1, 0, 0),
        create_test_case(0xC000, 0xC000, op_code, 0x8000, 0, 1, 0, 1),
        create_test_case(0x4000, 0x4001, op_code, 0x8001, 0, 1, 1, 0),
        create_test_case(0x0000, 0x0000, op_code, 0x0000, 1, 0, 0, 0),
        create_test_case(0x0001, 0xFFFF, op_code, 0x0000, 1, 0, 0, 1),
        create_test_case(0x8000, 0x8000, op_code, 0x0000, 1, 0, 1, 1),
    )
    # Clear flags as they will be affected by the tests
    alu.reset()
    check_test_cases(test_cases)


def test_subtract():
    op_code = 0b0101
    test_cases = (
        # Test all possible flag combinations
        create_test_case(0x0001, 0xFFFF, op_code, 0x0002, 0, 0, 0, 0),
        create_test_case(0x0001, 0x0000, op_code, 0x0001, 0, 0, 0, 1),
        create_test_case(0x0005, 0x0003, op_code, 0x0002, 0, 0, 0, 1),
        create_test_case(0x8000, 0x0001, op_code, 0x7FFF, 0, 0, 1, 1),
        create_test_case(0x0005, 0x000A, op_code, 0xFFFB, 0, 1, 0, 0),
        create_test_case(0xFFFF, 0x0001, op_code, 0xFFFE, 0, 1, 0, 1),
        create_test_case(0xA093, 0x0000, op_code, 0xA093, 0, 1, 0, 1),
        create_test_case(0x4000, 0xBFFF, op_code, 0x8001, 0, 1, 1, 0),
        create_test_case(0x0000, 0x0000, op_code, 0x0000, 1, 0, 0, 1),
        create_test_case(0x0032, 0x0032, op_code, 0x0000, 1, 0, 0, 1),
    )
    alu.reset()
    check_test_cases(test_cases)


def test_and():
    op_code = 0b0110
    # Set the o- and c-flag values of the ALU and make sure that they are
    # not affected
    alu.reset()
    alu._o_value = 1
    alu._c_value = 1
    test_cases = (
        create_test_case(0x0F0F, 0xF00F, op_code, 0x000F, 0, 0),
        create_test_case(0x80DA, 0xF000, op_code, 0x8000, 0, 1),
        create_test_case(0x0000, 0x0000, op_code, 0x0000, 1, 0),
        create_test_case(0xFFFF, 0x0000, op_code, 0x0000, 1, 0),
    )
    check_test_cases(test_cases)


def test_or():
    op_code = 0b0111
    # Set the o- and c-flag values of the ALU and make sure that they are
    # not affected
    alu.reset()
    alu._o_value = 1
    alu._c_value = 1
    test_cases = (
        create_test_case(0x0017, 0x0000, op_code, 0x0017, 0, 0),
        create_test_case(0x0000, 0x0385, op_code, 0x0385, 0, 0),
        create_test_case(0xAAAA, 0x5555, op_code, 0xFFFF, 0, 1),
        create_test_case(0x0000, 0x0000, op_code, 0x0000, 1, 0),
    )
    check_test_cases(test_cases)


def test_add_no_flags():
    op_code = 0b1000
    # Set the flag values of the ALU and make sure that they are not affected
    alu.reset()
    alu._z_value = 1
    alu._n_value = 1
    alu._o_value = 1
    alu._c_value = 1
    test_cases = (
        # Test cases that hit all possible flag combinations during normal add
        create_test_case(0x0004, 0x0008, op_code, 0x000C),
        create_test_case(0xC000, 0x4001, op_code, 0x0001),
        create_test_case(0x8000, 0x8001, op_code, 0x0001),
        create_test_case(0x8000, 0x0001, op_code, 0x8001),
        create_test_case(0xC000, 0xC000, op_code, 0x8000),
        create_test_case(0x4000, 0x4001, op_code, 0x8001),
        create_test_case(0x0000, 0x0000, op_code, 0x0000),
        create_test_case(0x0001, 0xFFFF, op_code, 0x0000),
        create_test_case(0x8000, 0x8000, op_code, 0x0000),
    )
    check_test_cases(test_cases)


def test_shift_left():
    # 16-bit logical left shift, AR only
    op_code = 0b1001
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        create_test_case(0x0006, 0x0000, op_code, 0x000C, 0, 0, None, 0),
        create_test_case(0x0006, 0xF593, op_code, 0x000C, 0, 0, None, 0),
        create_test_case(0x8001, 0x0000, op_code, 0x0002, 0, 0, None, 1),
        create_test_case(0x4001, 0x0000, op_code, 0x8002, 0, 1, None, 0),
        create_test_case(0xC001, 0x0000, op_code, 0x8002, 0, 1, None, 1),
        create_test_case(0x0000, 0x0000, op_code, 0x0000, 1, 0, None, 0),
        create_test_case(0x0000, 0x0256, op_code, 0x0000, 1, 0, None, 0),
    )
    check_test_cases(test_cases)

    # 32-bit logical left shift, ARHR
    op_code = 0b1010
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        create_hr_test_case(0x0000, 0x0001, op_code, 0x0000, 0x0002, 0, 0, 0),
        create_hr_test_case(0x0001, 0x0000, op_code, 0x0002, 0x0000, 0, 0, 0),
        create_hr_test_case(0x0001, 0x0001, op_code, 0x0002, 0x0002, 0, 0, 0),
        create_hr_test_case(0x8000, 0x8000, op_code, 0x0001, 0x0000, 0, 0, 1),
        create_hr_test_case(0x4000, 0x8000, op_code, 0x8001, 0x0000, 0, 1, 0),
        create_hr_test_case(0xC000, 0x8000, op_code, 0x8001, 0x0000, 0, 1, 1),
        create_hr_test_case(0x0000, 0x0000, op_code, 0x0000, 0x0000, 1, 0, 0),
        create_hr_test_case(0x8000, 0x0000, op_code, 0x0000, 0x0000, 1, 0, 1),
    )
    check_test_cases(test_cases)


def test_arithmetic_shift_right():
    # 16-bit arithmetic right shift, AR only
    op_code = 0b1011
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        create_test_case(0x0002, 0x0000, op_code, 0x0001, 0, 0, None, 0),
        create_test_case(0x0002, 0x0934, op_code, 0x0001, 0, 0, None, 0),
        create_test_case(0x0003, 0x0000, op_code, 0x0001, 0, 0, None, 1),
        create_test_case(0x8000, 0x0000, op_code, 0xC000, 0, 1, None, 0),
        create_test_case(0x8001, 0x0000, op_code, 0xC000, 0, 1, None, 1),
        create_test_case(0x0000, 0x0000, op_code, 0x0000, 1, 0, None, 0),
        create_test_case(0x0001, 0x0000, op_code, 0x0000, 1, 0, None, 1),
    )
    check_test_cases(test_cases)

    # 32-bit arithmetic right shift, ARHR
    op_code = 0b1100
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        # Focus on AR part only
        create_hr_test_case(0x0002, 0x0000, op_code, 0x0001, 0x0000, 0, 0, 0),
        # Focus on HR part only
        create_hr_test_case(0x0000, 0x0002, op_code, 0x0000, 0x0001, 0, 0, 0),
        # Both AR and HR are non-zero
        create_hr_test_case(0x400A, 0x0F06, op_code, 0x2005, 0x0783, 0, 0, 0),
        create_hr_test_case(0x000B, 0x0006, op_code, 0x0005, 0x8003, 0, 0, 0),
        # Check all possible combinations of flags
        create_hr_test_case(0x0003, 0x0000, op_code, 0x0001, 0x8000, 0, 0, 0),
        create_hr_test_case(0x0003, 0x0003, op_code, 0x0001, 0x8001, 0, 0, 1),
        create_hr_test_case(0x8000, 0x0000, op_code, 0xC000, 0x0000, 0, 1, 0),
        create_hr_test_case(0x800A, 0x8006, op_code, 0xC005, 0x4003, 0, 1, 0),
        create_hr_test_case(0xC005, 0x0003, op_code, 0xE002, 0x8001, 0, 1, 1),
        create_hr_test_case(0x0000, 0x0000, op_code, 0x0000, 0x0000, 1, 0, 0),
        create_hr_test_case(0x0000, 0x0001, op_code, 0x0000, 0x0000, 1, 0, 1),
    )
    check_test_cases(test_cases)


def test_logical_shift_right():
    op_code = 0b1101
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        create_test_case(0x0002, 0x0000, op_code, 0x0001, 0, 0, None, 0),
        create_test_case(0x8002, 0x9654, op_code, 0x4001, 0, 0, None, 0),
        create_test_case(0x0005, 0xA634, op_code, 0x0002, 0, 0, None, 1),
        create_test_case(0x8001, 0x0000, op_code, 0x4000, 0, 0, None, 1),
        create_test_case(0x0000, 0x45F2, op_code, 0x0000, 1, 0, None, 0),
        create_test_case(0x0001, 0x3D9E, op_code, 0x0000, 1, 0, None, 1),
    )
    check_test_cases(test_cases)


def test_rotate_left():
    # 16-bit left rotate, AR input only
    op_code = 0b1110
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        create_test_case(0x0001, 0x0000, op_code, 0x0002, 0, 0, None, 0),
        create_test_case(0x0008, 0x2971, op_code, 0x0010, 0, 0, None, 0),
        create_test_case(0x8001, 0x0952, op_code, 0x0003, 0, 0, None, 1),
        create_test_case(0x5020, 0xAECD, op_code, 0xA040, 0, 1, None, 0),
        create_test_case(0xC001, 0xF408, op_code, 0x8003, 0, 1, None, 1),
        create_test_case(0x0000, 0x1987, op_code, 0x0000, 1, 0, None, 0),
    )
    check_test_cases(test_cases)

    # 32-bit left rotate, ARHR input
    op_code = 0b1111
    # Set the o-flag value of the ALU and make sure that it is not affected
    alu.reset()
    alu._o_value = 1
    test_cases = (
        create_hr_test_case(0x0410, 0x0000, op_code, 0x0820, 0x0000, 0, 0, 0),
        create_hr_test_case(0x0000, 0x0410, op_code, 0x0000, 0x0820, 0, 0, 0),
        create_hr_test_case(0x0080, 0x8410, op_code, 0x0101, 0x0820, 0, 0, 0),
        create_hr_test_case(0xA000, 0x8001, op_code, 0x4001, 0x0003, 0, 0, 1),
        create_hr_test_case(0x7001, 0xC001, op_code, 0xE003, 0x8002, 0, 1, 0),
        create_hr_test_case(0xE021, 0xC001, op_code, 0xC043, 0x8003, 0, 1, 1),
        create_hr_test_case(0x0000, 0x0000, op_code, 0x0000, 0x0000, 1, 0, 0),
    )
    check_test_cases(test_cases)
