from simudator.processor.mia.modules.alu import ALU
from simudator.processor.mia.modules.alu_arithmetics import (
    add_binary_numbers,
    arithmetic_right_shift,
    bin_to_int,
    bin_to_int_twocomp,
    binary_and,
    binary_or,
    bitwise_not,
    int_to_bin,
    left_rotate,
    logical_left_shift,
    logical_right_shift,
)


def create_unary_test_case(op, res, carry=None):
    """
    Create a test case for a unary arithmetic function.

    Parameters
    ----------
    op : int
        Operand to the arithmetic function.
    res : int
        Expected result from the arithmetic function.
    carry : int
        Expected carry out from the function if any. Set it to ``None`` to
        signify that the function does not return a carry out bit.

    Returns
    -------
    tuple[int, int, int]
        Test case for a unary arithmetic function.
    """
    return op, res, carry


def check_unary_test(test_case, fun):
    """
    Assert that a unary arithmetic function passes a test case.

    Parameters
    ----------
    test_case : tuple[int, int, int]
        Test case for the unary arithmetic function.
    fun : callable
        Unary arithmetic function to test. Should operate on binary numbers
        represented as lists of bits.
        ``fun(list[int]) -> list[int] | tuple[list[int], int]``
    """
    (op, res, carry) = test_case
    bin_op = int_to_bin(op, ALU.WORD_LENGTH)
    bin_res = int_to_bin(res, ALU.WORD_LENGTH)
    if carry is None:
        assert fun(bin_op) == bin_res
    else:
        assert fun(bin_op) == (bin_res, carry)


def check_unary_tests(test_cases, fun):
    """
    Assert that a unary arithmetic function passes some test cases.

    Parameters
    ----------
    test_cases : list[tuple[int, int, int]]
        List of test cases for the unary arithmetic function.
    fun : callable
        Unary arithmetic function to test. Should operate on binary numbers
        represented as lists of bits.
        ``fun(list[int]) -> list[int] | tuple[list[int], int]``
    """
    for case in test_cases:
        check_unary_test(case, fun)


def create_binary_test_case(op_1, op_2, res, carry=None):
    """
    Create a test case for a binary arithmetic function.

    Parameters
    ----------
    op_1 : int
        First operand to the arithmetic function.
    op_2 : int
        Second operand to the arithmetic function.
    res : int
        Expected result from the arithmetic function.
    carry : int
        Expected carry out from the function if any. Set it to ``None`` to
        signify that the function does not return a carry out bit.

    Returns
    -------
    tuple[int, int, int, int]
        Test case for a binary arithmetic function.
    """
    return op_1, op_2, res, carry


def check_binary_test(test_case, fun):
    """
    Assert that a binary arithmetic function passes a test case.

    Parameters
    ----------
    test_case : tuple[int, int, int, int]
        Test case for the binary arithmetic function.
    fun : callable
        Binary arithmetic function to test. Should operate on binary numbers
        represented as lists of bits.
        ``fun(list[int], list[int]) -> list[int] | tuple[list[int], int]``
    """
    (op_1, op_2, res, carry) = test_case
    bin_op_1 = int_to_bin(op_1, ALU.WORD_LENGTH)
    bin_op_2 = int_to_bin(op_2, ALU.WORD_LENGTH)
    bin_res = int_to_bin(res, ALU.WORD_LENGTH)
    if carry is None:
        assert fun(bin_op_1, bin_op_2) == bin_res
    else:
        assert fun(bin_op_1, bin_op_2) == (bin_res, carry)


def check_binary_tests(test_cases, fun):
    """
    Assert that a binary arithmetic function passes some test cases.

    Parameters
    ----------
    test_case : list[tuple[int, int, int, int]]
        List of test cases for the binary arithmetic function.
    fun : callable
        Binary arithmetic function to test. Should operate on binary numbers
        represented as lists of bits.
        ``fun(list[int], list[int]) -> list[int] | tuple[list[int], int]``
    """
    for case in test_cases:
        check_binary_test(case, fun)


def test_int_binary_converter():
    # test int to binary
    word_length = ALU.WORD_LENGTH
    assert int_to_bin(0, word_length) == [0 for _ in range(word_length)]
    assert int_to_bin(1, word_length) == [0] * (word_length - 1) + [1]
    assert int_to_bin(32767, word_length) == [0] + [1] * (word_length - 1)
    assert int_to_bin(9, word_length) == [0] * (word_length - 4) + [1, 0, 0, 1]

    # test binary to int
    assert bin_to_int(int_to_bin(15, word_length)) == 15
    assert bin_to_int(int_to_bin(0, word_length)) == 0
    assert bin_to_int(int_to_bin(2**16, word_length)) == 2**16 - 1
    assert bin_to_int(int_to_bin(3576, word_length)) == 3576
    assert bin_to_int([0 for _ in range(word_length)]) == 0
    assert bin_to_int([0] * (word_length - 1) + [1]) == 1
    assert bin_to_int([1 for _ in range(word_length)]) == 2**16 - 1
    assert bin_to_int([0] + [1] * (word_length - 1)) == 32767
    assert bin_to_int([0] * (word_length - 4) + [1, 0, 0, 1]) == 9


def test_bitwise_not():
    test_cases = (
        create_unary_test_case(0, 2**16 - 1),
        create_unary_test_case(1, 2**16 - 1 - 1),
        create_unary_test_case(2, 2**16 - 1 - 2),
        create_unary_test_case(300, 2**16 - 1 - 300),
    )
    check_unary_tests(test_cases, bitwise_not)


def test_add():
    test_cases = (
        create_binary_test_case(0b01, 0b10, 0b11, 0),
        create_binary_test_case(0b00101, 0b10101, 0b11010, 0),
        create_binary_test_case(0, 0, 0, 0),
        create_binary_test_case(1, 2**16 - 1, 0, 1),
    )
    check_binary_tests(test_cases, add_binary_numbers)


def test_and():
    test_cases = (
        create_binary_test_case(0b01, 0b10, 0b00),
        create_binary_test_case(0b1, 0b1, 0b1),
        create_binary_test_case(0b11111, 0b00000, 0b00000),
        create_binary_test_case(0b11111, 0b10101, 0b10101),
    )
    check_binary_tests(test_cases, binary_and)


def test_or():
    test_cases = (
        create_binary_test_case(0b01, 0b10, 0b11),
        create_binary_test_case(0b000, 0b000, 0b00),
        create_binary_test_case(0b11111, 0b00000, 0b11111),
        create_binary_test_case(0b01010, 0b10101, 0b11111),
        create_binary_test_case(0b01110, 0b10001, 0b11111),
    )
    check_binary_tests(test_cases, binary_or)


def test_logical_left_shift():
    test_cases = (
        create_unary_test_case(0b01, 0b10, 0),
        create_unary_test_case(0b11, 0b110, 0),
        create_unary_test_case(0b0, 0b0, 0),
        create_unary_test_case(0b1000000000000000, 0b0, 1),
    )
    check_unary_tests(test_cases, logical_left_shift)


def test_arithmetic_right_shift():
    test_cases = (
        create_unary_test_case(0b0, 0b0, 0),
        create_unary_test_case(0b1, 0b0, 1),
        create_unary_test_case(0b10, 0b1, 0),
        create_unary_test_case(0b1111111111111111, 0b1111111111111111, 1),
        create_unary_test_case(0b0111111111111110, 0b0011111111111111, 0),
    )
    check_unary_tests(test_cases, arithmetic_right_shift)


def test_logical_right_shift():
    test_cases = (
        create_unary_test_case(0b0, 0b0, 0),
        create_unary_test_case(0b1, 0b0, 1),
        create_unary_test_case(0b10, 0b01, 0),
        create_unary_test_case(0b1101, 0b0110, 1),
    )
    check_unary_tests(test_cases, logical_right_shift)


def test_rotate_left():
    test_cases = (
        create_unary_test_case(0b1000000000000000, 0b0000000000000001, 1),
        create_unary_test_case(0b0, 0b0, 0),
        create_unary_test_case(0b01, 0b10, 0),
    )
    check_unary_tests(test_cases, left_rotate)
