from unittest.mock import mock_open, patch

from simudator.core.modules import Flag
from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.lc import LC


def test_default_name():
    cpu = Processor()
    s = Signal(cpu)
    lc = LC(s, s, s, s)
    assert lc.name == "LC"


def test_name():
    cpu = Processor()
    s = Signal(cpu)
    lc = LC(s, s, s, s, "lc")
    assert lc.name == "lc"


def test_read_from_bus():
    cpu = Processor()
    mM_control_s = Signal(cpu)
    bus_input_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s)

    cpu.add_module(lc)

    bus_input_s.update_value(10)
    mM_control_s.update_value(2)

    cpu.do_tick()
    assert lc._value == 10


def test_read_from_uADR():
    cpu = Processor()
    mM_control_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    bus_input_s = Signal(cpu)
    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s)

    cpu.add_module(lc)

    mM_uADR_s.update_value(10)
    mM_control_s.update_value(3)

    cpu.do_tick()
    assert lc._value == 10


def test_write_to_l_flag():
    cpu = Processor()

    # signals needed
    mM_control_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    l_flag_output_s = Signal(cpu)
    bus_input_s = Signal(cpu)

    # modules needed to run the test
    # initialize loop counter to 1, so when it is decremented by
    # one it reaches zero and sets the L flag
    # initialize the L flag as a register with 1 bit length and
    # the value 0
    lc_value = 1
    l_bit_length = 1
    l_value = 0
    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s, "lc", lc_value)
    l_flag = Flag(l_flag_s, l_flag_output_s, l_bit_length, l_value)

    # add modules to the processor
    cpu.add_module(lc)
    cpu.add_module(l_flag)

    # update the value of the control signal and the micro controller
    mM_uADR_s.update_value(0)
    mM_control_s.update_value(0b01)

    # The loop counter will now change its value from 1 to 0 which
    # should set the L flag to 1 in the same cycle
    cpu.do_tick()
    assert l_flag._value == 1


def test_reset_l_flag():
    cpu = Processor()

    mM_control_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    l_flag_output_s = Signal(cpu)
    bus_input_s = Signal(cpu)

    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s, "lc", 1)
    l_flag = Flag(l_flag_s, l_flag_output_s, 8)

    cpu.add_module(lc)
    cpu.add_module(l_flag)

    mM_uADR_s.update_value(0)
    mM_control_s.update_value(3)

    cpu.do_tick()

    mM_uADR_s.update_value(10)

    cpu.do_tick()
    assert l_flag._value == 0


def test_lc_do_nothing():
    cpu = Processor()

    mM_control_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    bus_input_s = Signal(cpu)

    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s, "lc", value=4)

    cpu.add_module(lc)

    # tell lc to decrement by one
    mM_control_s.update_value(0b01)

    cpu.do_tick()
    assert lc._value == 3

    # tell lc to do nothing
    mM_control_s.update_value(0b00)

    cpu.do_tick()
    assert lc._value == 3


def test_get_state():
    # init all the modules
    cpu = Processor()
    bus_input_s = Signal(cpu)
    mM_control_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s)

    mM_uADR_s.update_value(10)
    bus_input_s.update_value(100)
    cpu.add_module(lc)

    mM_control_s.update_value(0b01)
    cpu.do_tick()

    state = lc.get_state()
    assert state["name"] == "LC"
    assert state["value"] == 255
    assert state["read_from_bus"] is False
    assert state["read_from_uADR"] is False
    assert state["decrement_by_one"] is True

    mM_control_s.update_value(0b10)
    cpu.do_tick()

    state = lc.get_state()
    assert state["name"] == "LC"
    assert state["value"] == 100
    assert state["read_from_bus"] is True
    assert state["read_from_uADR"] is False
    assert state["decrement_by_one"] is False

    mM_control_s.update_value(0b11)
    cpu.do_tick()

    state = lc.get_state()
    assert state["name"] == "LC"
    assert state["value"] == 10
    assert state["read_from_bus"] is False
    assert state["read_from_uADR"] is True
    assert state["decrement_by_one"] is False


def test_get_parameter():
    # init all the modules
    cpu = Processor()
    bus_input_s = Signal(cpu)
    mM_control_s = Signal(cpu)
    mM_uADR_s = Signal(cpu)
    l_flag_s = Signal(cpu)
    lc = LC(mM_control_s, bus_input_s, l_flag_s, mM_uADR_s)

    parameter = lc.get_parameter()
    assert parameter["bit_length"] == 8


def test_save_state():
    cpu = Processor()
    s = Signal(cpu)
    lc = LC(s, s, s, s)

    dummy_file_path = "dummy"

    mock_file = mock_open()

    content = lc.name + ":\n"
    content += "value: " + hex(lc._value)[2:] + "\n\n"
    with patch("builtins.open", mock_file):
        lc.save_state_to_file(dummy_file_path)
        mock_file.assert_called_once_with(dummy_file_path, "a")
        mock_file().write.assert_called_once_with(content)
