from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_grx import GRX
from unittest.mock import patch, mock_open


def test_default_name():
    cpu = Processor()
    s = Signal(cpu)
    bus_id = 5
    grx = GRX(s, s, s, s, s, s, bus_id)

    assert grx.name == "GRx"


def test_name():
    cpu = Processor()
    s = Signal(cpu)
    bus_id = 5
    grx = GRX(s, s, s, s, s, s, bus_id, "grx")

    assert grx.name == "grx"


def test_read_from_bus():
    cpu = Processor()

    # signals needed
    to_bus = Signal(cpu)
    from_bus = Signal(cpu)
    bus_control = Signal(cpu)
    grx_control = Signal(cpu)
    m_control = Signal(cpu)
    s_control = Signal(cpu)

    bus_id = 1

    grx = GRX(to_bus, from_bus, bus_control, grx_control, m_control, s_control, bus_id)

    cpu.add_module(grx)

    # set the cpu to read from bus
    bus_control.update_value((0, 1))
    # set the value to be read to 7
    from_bus.update_value(0b111)
    # set the s bit to index using grx
    s_control.update_value(0)
    # write to the 4th register
    grx_control.update_value(0b11)

    cpu.do_tick()
    assert grx.registers[3] == 7

    # index using the m bits
    s_control.update_value(1)
    # write to the 3rd register
    m_control.update_value(0b10)

    cpu.do_tick()
    assert grx.registers[2] == 7


def test_write_to_bus():
    cpu = Processor()

    # signals needed
    to_bus = Signal(cpu)
    from_bus = Signal(cpu)
    bus_control = Signal(cpu)
    grx_control = Signal(cpu)
    m_control = Signal(cpu)
    s_control = Signal(cpu)

    bus_id = 1

    # give registers none default values
    registers = [1, 2, 3, 4]

    grx = GRX(
        to_bus,
        from_bus,
        bus_control,
        grx_control,
        m_control,
        s_control,
        bus_id,
        "grx",
        registers,
    )

    cpu.add_module(grx)

    # set the control so that grx should not read nor write to/from bus
    # grx should output 'None'
    bus_control.update_value((3, 2))

    cpu.do_tick()
    assert to_bus.get_value() is None

    # set the grx to write to the bus
    bus_control.update_value((1, 3))
    # set the registers to be indexed by the grx bits
    s_control.update_value(0)
    # set the grx to select the first register
    grx_control.update_value(0b00)

    cpu.do_tick()
    assert to_bus.get_value() == 1

    # set the registers to be indexed by the m bits
    s_control.update_value(1)
    # set the m to select the second register
    m_control.update_value(0b01)

    cpu.do_tick()
    assert to_bus.get_value() == 2


def test_set_value_from_bus():
    cpu = Processor()

    # signals needed
    to_bus = Signal(cpu)
    from_bus = Signal(cpu)
    bus_control = Signal(cpu)
    grx_control = Signal(cpu)
    m_control = Signal(cpu)
    s_control = Signal(cpu)

    bus_id = 1

    grx = GRX(to_bus, from_bus, bus_control, grx_control, m_control, s_control, bus_id)

    cpu.add_module(grx)

    # 0 and 1 are valid values, all else should throw value error
    s_control.update_value(2)

    bus_control.update_value((2, 1))

    error = False

    try:
        cpu.do_tick()
    except ValueError:
        error = True

    assert error is True

def test_save_state():
    cpu = Processor()
    s = Signal(cpu)
    grx = GRX(s, s, s, s, s, s, 1)

    dummy_file_path = "dummy"

    mock_file = mock_open()

    content = grx.name + ":\n"
    for index, value in enumerate(grx.registers):
        content += str(hex(index)[2:]) + ": "
        content += str(hex(value)[2:])
        content += "\n"

    content += "\n"
    with patch("builtins.open", mock_file):
        grx.save_state_to_file(dummy_file_path)
        mock_file.assert_called_once_with(dummy_file_path, "a")
        mock_file().write.assert_called_once_with(content)
