from unittest.mock import mock_open, patch

from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.micro_pc import MicroPC


def test_upc():
    cpu = Processor()
    control_signal = Signal(cpu)
    f_k1 = Signal(cpu)
    f_k2 = Signal(cpu)
    t_supc = Signal(cpu, value=-1)
    f_supc = Signal(cpu)
    t_um = Signal(cpu)
    f_um = Signal(cpu)
    upc = MicroPC(control_signal, f_k1, f_k2, t_supc, f_supc, t_um, f_um)
    cpu.add_module(upc)

    control_signal.update_value(0b000)
    f_k1.update_value(10)
    f_k2.update_value(20)
    f_supc.update_value(30)
    f_um.update_value(40)

    # Test uPC += 1 with control signal 0b000
    cpu.do_tick()
    assert upc._value == 1
    assert t_supc.get_value() == -1
    assert t_um.get_value() == upc._value

    # Test uPC = K1 with control signal 0b001
    control_signal.update_value(0b001)
    cpu.do_tick()
    assert upc._value == f_k1.get_value()
    assert t_supc.get_value() == -1
    assert t_um.get_value() == upc._value

    # Test uPC = K2 with control signal 0b010
    control_signal.update_value(0b010)
    cpu.do_tick()
    assert upc._value == f_k2.get_value()
    assert t_supc.get_value() == -1
    assert t_um.get_value() == upc._value

    # Test uPC = 0 with control signal 0b011
    control_signal.update_value(0b011)
    cpu.do_tick()
    assert upc._value == 0
    assert t_supc.get_value() == -1
    assert t_um.get_value() == upc._value

    # Test uPC = uADR from uM with control signal 0b100
    control_signal.update_value(0b100)
    cpu.do_tick()
    assert upc._value == f_um.get_value()
    assert t_supc.get_value() == -1
    assert t_um.get_value() == upc._value

    # Test uPC = SuPC with control signal 0b101
    control_signal.update_value(0b101)
    cpu.do_tick()
    assert upc._value == f_supc.get_value()
    assert t_supc.get_value() == -1
    assert t_um.get_value() == upc._value

    # Test SuPC = uPC + 1, uPC = uADR from uM with control signal 0b110
    control_signal.update_value(0b110)
    temp = upc._value + 1
    cpu.do_tick()
    assert upc._value == f_um.get_value()
    assert t_supc.get_value() == temp
    assert t_um.get_value() == upc._value


def test_save_state():
    cpu = Processor()
    s = Signal(cpu)
    upc = MicroPC(s, s, s, s, s, s, s)

    dummy_file_path = "dummy"

    mock_file = mock_open()

    content = upc.name + ":\n"
    content += "value: " + hex(upc._value)[2:] + "\n\n"
    with patch("builtins.open", mock_file):
        upc.save_state_to_file(dummy_file_path)
        mock_file.assert_called_once_with(dummy_file_path, "a")
        mock_file().write.assert_called_once_with(content)
